import { uri, CarePlan, Identifier, Reference, Period, canonical, dateTime, CarePlanStatus, CarePlanIntent, CarePlanActivity, id, Resource } from 'fhir';
import { DomainResourceBuilder } from './domain-resource-builder';
/**
 * Class representing a FHIR CarePlan builder
 */
export declare class CarePlanBuilder extends DomainResourceBuilder {
    /**
     * The CarePlan that is being built
     * @protected
     * @type { CarePlan }
     * @ignore
     */
    protected resource: CarePlan;
    /**
     * Creates a new FHIR CarePlanBuilder
     */
    constructor();
    /**
     * Sets a copy of an existing carePlan for this CarePlanBuilder to allow for manipulating, or interacting with, the
     * carePlan through this CarePlanBuilder
     * If no argument is provided, a new CarePlan will be generated
     * @param { CarePlan } carePlan - The carePlan to set
     * @returns { CarePlanBuilder } The builder
     * @override
     */
    use(carePlan?: CarePlan): this;
    /**
     * Adds the resource provided as an argument to the resource's contained array, if a resource that is in conflict does not already exist in the array.
     * @param contained The resource to be added to the contained array
     * @returns { CarePlanBuilder } the Builder
     * @throws { AlreadyExistsError } If the resource already exists
     * @override
     */
    addContained(resource: Resource): this;
    /**
     * Returns either the identifier identified by the provided system from the identifier-array or the entire identifier array, if no system is provided
     * @returns { Identifier | undefined | Identifier[] } The identifier if it exists or the identifier-array
     */
    getIdentifier(): Identifier[];
    getIdentifier(system: uri): Identifier | undefined;
    /**
     * Replaces the identifier array with the provided array
     * If no argument is provided, the field will be removed
     * @param { Identifier[] } identifiers - the identifier to be set
     * @returns { CarePlanBuilder } - this CarePlanBuilder
     */
    setIdentifier(identifier?: Identifier[]): this;
    /**
     * Appends an identifier to the carePlan, if an identifier within the system does not already exist
     * @param { Identifier } identifier - the identifier to be appended
     * @returns { CarePlanBuilder } - this CarePlanBuilder
     * @throws { MissingPropertyError } When identifier is missing necessary properties
     * @throws { AlreadyExistsError } When an identifier in that system already exists in the CarePlan
     */
    addIdentifier(identifier: Identifier): this;
    /**
     * Returns all plan definition canonical references for carePlan
     * @returns { canonical[] } - the canonicals
     */
    getInstantiatesCanonical(): canonical[];
    /**
     * Replaces the instantiatesCanonical array with the provided array
     * If no argument is provided, the field will be removed
     * @param { canonical[] } canonical - the array to be set
     * @returns { CarePlanBuilder } - this CarePlanBuilder
     */
    setInstantiatesCanonical(canonical?: canonical[]): this;
    /**
     * Appends a plan definition canonical reference to the carePlan, if it does not already exist
     * @param { canonical } canonical - the canonical reference to be appended
     * @returns { CarePlanBuilder } - this CarePlanBuilder
     * @throws { AlreadyExistsError } When the canonical already exists on the CarePlan
     */
    addInstantiatesCanonical(canonical: canonical): this;
    /**
     * Returns the carePlan's status field
     * @returns { code } - the status of the carePlan
     */
    getStatus(): CarePlanStatus;
    /**
     * Sets the carePlan's status field
     * If no argument is provided, the field will be removed
     * @param { CarePlanStatus } status - the status to be set. Must be one of: draft | active | suspended | completed | entered-in-error | cancelled | unknown
     * @returns { CarePlanBuilder } - The builder
     */
    setStatus(status: CarePlanStatus): this;
    /**
     * Returns the carePlan's intent field. This field is contrained to 'order'
     * @returns { intent } - the intent of the carePlan
     */
    getIntent(): CarePlanIntent;
    /**
     * Sets the carePlan's intent-field
     * If no argument is provided, the field will be removed
     * @param { CarePlanIntent } intent - the intent to be set
     * @returns { CarePlanBuilder } - The builder
     */
    setIntent(intent: CarePlanIntent): this;
    /**
     * Returns the carplan's title field
     * @return { string } the title of the carePlan
     */
    getTitle(): string | undefined;
    /**
     * Sets the carePlan's title field.
     * If no argument is provided, the field will be removed
     * @param { string } title - the title of the carePlan
     * @return { CarePlanBuilder } this CarePlanBuilder
     */
    setTitle(title?: string): this;
    /**
     * Returns the carePlan's subject
     * @return { Reference } subject - the subject of the carePlan
     */
    getSubject(): Reference;
    /**
     * Sets the carePlan's subject field
     * If no argument is provided, the field will be removed
     * @param { Reference } subject - the subject which the carePlan refereces
     * @return { CarePlanBuilder } this CarePlanBuilder
     */
    setSubject(subject?: Reference): this;
    /**
     * Returns the carePlan's period
     * @returns { Period } period - the period within which the carePlan is active
     */
    getPeriod(): Period | undefined;
    /**
     * Sets the carePlan's period field
     * If no argument is provided, the field will be removed
     * @param { Period } period - the period within which the carePlan is active
     * @return { CarePlanBuilder } this CarePlanBuilder
     */
    setPeriod(period?: Period): this;
    /**
     * Returns the carePlan's encounter
     * @returns { Reference } encounter - the encounter to which the carePlan is related.
     */
    getEncounter(): Reference;
    /**
     * Sets the carePlan's encounter field
     * If no argument is provided, the field will be removed
     * @param { Reference } encounter - the encounter to which the carePlan is related
     * @return { CarePlanBuilder } this CarePlanBuilder
     */
    setEncounter(encounter?: Reference): this;
    /**
     * Returns the carePlan's created field
     * If no argument is provided, the field will be removed
     * @returns { dateTime } created - the date the carePlan's was created
     */
    getCreated(): dateTime;
    /**
     * Sets the carePlan's created field
     * If no argument is provided, the field will be removed
     * @param { dateTime } created - the date the carePlan's was created
     * @returns { CarePlanBuilder } this CarePlanBuilder
     */
    setCreated(created?: dateTime): this;
    /**
     * Returns the carePlan's author
     * @return { Reference } author - the author of the carePlan
     */
    getAuthor(): Reference;
    /**
     * Sets the carePlan's author field
     * If no argument is provided, the field will be removed
     * @param { Reference } author - the author which the carePlan refereces
     * @return { CarePlanBuilder } The builder
     */
    setAuthor(author?: Reference): this;
    /**
     * Returns an array of references to conditions, that this carePlan addresses
     * @return { Reference[] } - the array of references
     */
    getAddresses(): Reference[];
    /**
     * Replace the array of references to conditions, that this carePlan addresses with the provided array
     * If no argument is provided, the field will be removed
     * @param { Reference[] } addresses The array of references
     * @returns { CarePlanBuilder } - The builder
     */
    setAddresses(addresses?: Reference[]): this;
    /**
     * Appends a reference to a condition, that this carePlan addresses
     * @param { Reference } addresses - the reference to be appended
     * @returns { CarePlanBuilder } - this CarePlanBuilder
     * @throws { MissingPropertyError } When the Reference is missing necessary properties
     */
    addAddresses(reference: Reference): this;
    /**
     * Returns either the carePlanActivity identified by the provided id from the activity-array or the entire activity-array, if no id is provided
     * @returns { CarePlanActivity[] } The carePlanActivity identified by the provided id or the activity-array
     */
    getActivity(): CarePlanActivity[];
    getActivity(id: id): CarePlanActivity | undefined;
    /**
     * Sets the activity-array to the provided array
     * If no argument is provided, the field will be removed
     * @param { CarePlanActivity[] } activity The array to be set as the action-array
     * @returns the Builder
     */
    setActivity(activity?: CarePlanActivity[]): this;
    /**
     * Adds the activity provided as an argument to the resource's activity-array
     * @param { CarePlanActivity } activity The activity to be added to the activity-array
     * @returns the Builder
     * @throws { AlreadyExistsError } If the activity is in conflict with an acitivty already existing in the activity array
     */
    addActivity(activity: CarePlanActivity): this;
    /**
     * Returns whether the two provided resources are in conflict
     * @param { Resource } contained First provided resource
     * @param { Resource } resource second provided resource
     * @returns { boolean } Whether the two resources are in conflict
     * @ignore
     */
    private containedConflictsWith;
    /**
     * Returns the activity from the activity-array that is in conflict with the provided activity, if one exists
     * @param { CarePlanActivity } activity The activity to test for conflicts
     * @returns { CarePlanActivity } The activity that is in conflict with the provided activity
     *
     */
    getConflictingActivity(activity: CarePlanActivity): CarePlanActivity | undefined;
    /**
     * Returns the resource from the contained-array, that is in conflict with the provided resource, if one exists
     * @param { Resource } activity The activity to test for conflicts
     * @returns { Resource } The resource that is in conflict with the provided activity
     *
     */
    getConflictingContained(resource: Resource): Resource | undefined;
    /**
     * Creates an ISO-string for current time
     * @returns { string } An ISO-string for current time
     * @ignore
     */
    _generateIsoDate(): string;
    /**
     * Make a new object that implements a FHIR CarePlan
     * @returns { CarePlan } - an object that implements FHIR CarePlan
     * @ignore
     * @override
     */
    protected make(): CarePlan;
    /**
     * Builds and returns the carePlan
     * @returns { CarePlan } - the carePlan
     * @override
     */
    build(): CarePlan;
}
