import { Element, id, Extension, url } from 'fhir';
/**
 * Class representing a FHIR Element builder
 */
export declare class ElementBuilder {
    /**
     * The Element that is being built
     * @protected
     * @type { Element }
     * @ignore
     */
    protected element: Element;
    /**
     * Creates a new FHIR ElementBuilder
     */
    constructor();
    /**
     * Sets a copy of an existing element for this ElementBuilder to allow for manipulating, or interacting with, the
     * element through this ElementBuilder
     * If no argument is provided, a new Element will be generated
     * @param { Element } element - The element to set
     * @returns { ElementBuilder } The builder
     */
    use(element?: Element): this;
    /**
     * Returns the element's id
     * @returns { id } The element's id
     */
    getId(): id | undefined;
    /**
     * Sets the element's id to the id provided as an argument
     * If no argument is provided, the field will be removed
     * @param { id } id the id to be set
     * @returns the Builder
     */
    setId(id?: id): this;
    /**
     * Returns either the extensions identified by the provided url from the extension array or the entire extension array, if no url is provided
     * @returns { Resource[] } The extensions if they exists or the extension-array
     */
    getExtension(url?: url): Extension[];
    /**
     * Sets the resource'extension-array to the provided array
     * If no argument is provided, the field will be removed
     * @param { Extension[] } extension The array to be set as the resource's extension array
     * @returns the Builder
     */
    setExtension(extension?: Extension[]): this;
    /**
     * Adds the extension provided as an argument to the resource's extension array
     * @param extension The resource to be added to the extension array
     * @returns the Builder
     */
    addExtension(extension: Extension): this;
    /**
     * Returns whether or not the element provided as an argument is in conflict with the element used in the builder
     * @param { Element } element The element to test for conflict
     * @returns { boolean } Whether it is in conflict
     */
    conflictsWith(element: Element): boolean;
    /**
     * Make a new object that implements a FHIR Element
     * @returns { Element } - an object that implements FHIR Element
     */
    protected make(): Element;
    /**
     * Builds and returns the element
     * @returns { Element } - the element
     */
    build(): Element;
}
