import { BackboneElement, Extension, uri } from 'fhir'
import { ElementBuilder } from './element-builder'

/**
 * Class representing a FHIR Resource builder
 */
export class BackboneElementBuilder extends ElementBuilder {
  /**
   * The BackboneElement that is being built
   * @protected
   * @type { BackboneElement }
   * @ignore
   */
  protected element: BackboneElement

  /**
   * Creates a new FHIR BackboneElementBuilder
   */
  public constructor() {
    super()
    this.element = this.make()
  }

  /**
   * Sets a copy of an existing backboneElement for this builder to allow for manipulation
   * If no argument is provided, a new BackboneElement will be generated
   * @param { BackboneElement } backboneElement - the backboneElement to be set
   * @returns { BackboneElementBuilder } The builder
   * @override
   */
  public use(backboneElement?: BackboneElement): this {
    this.element = backboneElement ? { ...backboneElement } : this.make()
    return this
  }

  /**
   * Returns either the extensions identified by the provided url from the modifierExtension array or the entire modifierExtension array, if no url is provided
   * @param { uri } [url] The url of the extensions to retrieve (if applicable)
   * @returns { Resource | undefined | Resource[] } The extensions if it exists or the modifierExtension-array
   */
  public getModifierExtension(url?: uri): Extension[] {
    if (url) {
      return this.getModifierExtension().filter(
        extension => extension.url === url
      )
    }
    return this.element.modifierExtension || []
  }

  /**
   * Sets the resource's modifierExtension-array to the provided array. If no argument is provided, the field will be removed
   * @param { Extension[] } extensions The array to be set as the resource's modifierExtension array
   * @returns the Builder
   */
  public setModifierExtension(extension?: Extension[]): this {
    extension
      ? (this.element.modifierExtension = extension)
      : delete this.element.modifierExtension
    return this
  }

  /**
   * Adds the extension provided as an argument to the resource's modifierExtension array
   * @param extension The resource to be added to the modifierExtension array
   * @returns the Builder
   */
  public addModifierExtension(extension: Extension): this {
    this.setModifierExtension([...this.getModifierExtension(), extension])
    return this
  }

  /**
   * Make a new object that implements a FHIR BackboneElement
   * @returns { BackboneElement } An object that implements FHIR BackboneElement
   * @override
   */
  protected make(): BackboneElement {
    const backboneElement = {} as BackboneElement
    return backboneElement
  }

  /**
   * Builds and returns the backboneElement
   * @returns { BackboneElement } The backboneElement
   * @override
   */
  public build(): BackboneElement {
    return this.element
  }
}
