import { CarePlanActivity, CodeableConcept, Reference } from 'fhir'
import { BackboneElementBuilder } from './backbone-element-builder'
import { ReferenceBuilder } from './reference-builder'
import { ExtensionBuilder } from './extension-builder'
import { Constants } from './global/constants'

/**
 * Class representing a FHIR CarePlanActivity builder
 * FIXME: Incomplete
 */
export class CarePlanActivityBuilder extends BackboneElementBuilder {
  /**
   * The CarePlanActivity that is being built
   * @protected
   * @type { CarePlanActivity }
   * @ignore
   */
  protected element: CarePlanActivity

  /**
   * Creates a new FHIR CarePlanActivityBuilder
   */
  public constructor() {
    super()
    this.element = this.make()
  }

  /**
   * Sets a copy of an existing CarePlanActivity for this CarePlanActivityBuilder to allow for manipulating, or interacting with, the
   * carePlanActivity through this CarePlanActivityBuilder
   * If no argument is provided, a new CarePlanActivity will be generated
   * @param { CarePlanActivity } careplan - The carePlanActivity to set
   * @returns { CarePlanActivityBuilder } The builder
   * @override
   */
  public use(carePlanActivity?: CarePlanActivity): this {
    this.element = carePlanActivity ? { ...carePlanActivity } : this.make()
    return this
  }

  /**
   * Returns the outcomeCodeableConcept-field of the CarePlanActivity
   * @returns { CodeableConcept[] } The outcomeCodeableConcept-field
   */
  public getOutcomeCodeableConcept(): CodeableConcept[] {
    return this.element.outcomeCodeableConcept || []
  }

  /**
   * Sets the outcomeCodeableConcept-field of the CarePlanActivity with the provided CodeableConcept-array. 
   * If no argument is provided, the field will be removed
   * @returns { CarePlanActivityBuilder } The builder
   */
  public setOutcomeCodeableConcept(
    outcomeCodeableConcept?: CodeableConcept[]
  ): this {
    outcomeCodeableConcept
      ? (this.element.outcomeCodeableConcept = outcomeCodeableConcept)
      : delete this.element.outcomeCodeableConcept
    return this
  }

  /**
   * Adds a CodeableConcept to the outcomeCodeableConcept-array of the CarePlanActivity
   * @returns { CarePlanActivityBuilder } The builder
   */
  public addOutcomeCodeableConcept(
    outcomeCodeableConcept: CodeableConcept
  ): this {
    this.setOutcomeCodeableConcept([
      ...this.getOutcomeCodeableConcept(),
      outcomeCodeableConcept,
    ])
    return this
  }

  /**
   * Returns the outcomeReference-field of the CarePlanActivity
   * @returns { Reference[] } The outcomeReference-field
   */
  public getOutcomeReference(): Reference[] {
    return this.element.outcomeReference || []
  }

  /**
   * Sets the outcomeReference-field of the CarePlanActivity with the provided Reference-array
   * If no argument is provided, the field will be removed
   * @returns { CarePlanActivityBuilder } The builder
   */
  public setOutcomeReference(outcomeReference?: CodeableConcept[]): this {
    outcomeReference
      ? (this.element.outcomeReference = outcomeReference)
      : delete this.element.outcomeReference
    return this
  }

  /**
   * Adds a CodeableConcept to the outcomeCodeableConcept-array of the CarePlanActivity
   * @returns { CarePlanActivityBuilder } The builder
   */
  public addOutcomeReference(outcomeReference: Reference): this {
    this.setOutcomeReference([...this.getOutcomeReference(), outcomeReference])
    return this
  }

  /**
   * Returns the reference-field of the CarePlanActivity if it is set
   * @returns { Reference } The reference-field
   */
  public getReference(): Reference | undefined {
    return this.element.reference
  }

  /**
   * Sets the reference-field of the CarePlanActivity with the provided reference
   * If no argument is provided, the field will be removed
   * @returns { CarePlanActivityBuilder } The builder
   */
  public setReference(reference?: Reference): this {
    reference
      ? (this.element.reference = reference)
      : delete this.element.reference
    return this
  }

  /**
   * Returns whether or not the carePlanActivity provided as an argument is in conflict with the carePlanActivity used in the builder
   * @param { CarePlanActivity } activity The carePlanActivity to test for conflict
   * @returns { boolean } Whether it is in conflict
   * @override
   */
  public conflictsWith(activity: CarePlanActivity) {
    let conflict: boolean = super.conflictsWith(activity)
    const activityBuilder: CarePlanActivityBuilder = new CarePlanActivityBuilder().use(
      activity
    )

    if (
      this.getReference() &&
      activityBuilder.getReference() &&
      new ReferenceBuilder()
        .use(this.getReference())
        .conflictsWith(activityBuilder.getReference() as Reference)
    ) {
      conflict = true
    }

    const extension =
      this.getExtension(Constants.carePlan.thresholdSet)[0] || undefined
    if (!extension) return conflict
    const otherExtension =
      activityBuilder.getExtension(Constants.carePlan.thresholdSet)[0] ||
      undefined
    if (!otherExtension) return conflict

    const extensionBuilder: ExtensionBuilder = new ExtensionBuilder().use(
      extension
    )
    const otherExtensionBuilder: ExtensionBuilder = new ExtensionBuilder().use(
      otherExtension
    )

    if (
      extensionBuilder.getValueReference() &&
      otherExtensionBuilder.getValueReference() &&
      new ReferenceBuilder()
        .use(extensionBuilder.getValueReference())
        .conflictsWith(otherExtensionBuilder.getValueReference() as Reference)
    ) {
      conflict = true
    }
    return conflict
  }

  /**
   * Make a new object that implements a FHIR CarePlanActivity
   * @returns { CarePlanActivity } - an object that implements FHIR CarePlanActivity
   * @override
   */
  protected make(): CarePlanActivity {
    return {} as CarePlanActivity
  }

  /**
   * Builds and returns the carePlanActivity
   * @returns { CarePlanActivity } - the carePlanActivity
   * @override
   */
  public build(): CarePlanActivity {
    return this.element
  }
}
