import { DomainResource, Resource, id, Extension, uri, Narrative } from 'fhir'
import { ResourceBuilder } from './resource-builder'
import { AlreadyExistsError } from './exceptions'

/**
 * Class representing a FHIR DomainResource builder
 */
export class DomainResourceBuilder extends ResourceBuilder {
  /**
   * The DomainResource that is being built
   * @protected
   * @type { DomainResource }
   * @ignore
   */
  protected resource: DomainResource

  /**
   * Creates a new FHIR DomainResourceBuilder
   */
  public constructor() {
    super()
    this.resource = this.make()
  }

  /**
   * Sets a copy of an existing domainResource for this builder to allow for manipulation,
   * If no argument is provided, a new DomainResource will be generated
   * @param { DomainResource } domainResource - the domainResource to be set
   * @returns { DomainResourceBuilder } The builder
   * @override
   */
  public use(domainResource?: DomainResource): this {
    this.resource = domainResource ? { ...domainResource } : this.make()
    return this
  }

  /**
   * Returns the text-field of the DomainResource
   * @returns The text-field
   */
  public getText(): Narrative | undefined {
    return this.resource.text
  }

  /**
   * Sets the text-field of the DomainResource
   * If no argument is provided, the field will be removed
   * @param { string } text - The value to be set in the text-field
   * @returns { DomainResourceBuilder } - The builder
   */
  public setText(text?: Narrative): this {
    text ? (this.resource.text = text) : delete this.resource.text
    return this
  }

  /**
   * Returns either the resource identified by the provided id from the contained array or the entire contained array, if no id is provided
   * @returns { Resource | undefined | Resource[] } The resource if it exists or the contained-array
   */
  public getContained(): Resource[]
  public getContained(id: id): Resource | undefined
  public getContained(id?: id): Resource | undefined | Resource[] {
    if (id) {
      // If argument is local reference, remove '#'
      id = id.substr(0, 1) === '#' ? id.substr(1, id.length - 1) : id
      return this.getContained().find(resource => resource.id === id)
    }
    return this.resource.contained || []
  }

  /**
   * Sets the resource's contained-array to the provided array
   * If no argument is provided, the field will be removed
   * @param { Resource[] } contained The array to be set as the resource's contained array
   * @returns the Builder
   */
  public setContained(contained?: Resource[]): this {
    contained
      ? (this.resource.contained = contained)
      : delete this.resource.contained
    return this
  }

  /**
   * Adds the resource provided as an argument to the resource's contained array, if a resource with that id does not already exist in the array.
   * @param contained The resource to be added to the contained array
   * @returns the Builder
   * @throws { AlreadyExistsError } If the resource already exists
   */
  public addContained(contained: Resource): this {
    if (
      this.getContained().find(resource =>
        new ResourceBuilder().use(resource).conflictsWith(contained)
      )
    ) {
      throw new AlreadyExistsError(
        'A resource with that id already exists in the contained-array'
      )
    }
    this.setContained([...this.getContained(), contained])
    return this
  }

  /**
   * Returns either the extensions identified by the provided url from the extension array or the entire extension array, if no url is provided
   * @param { uri } [url] The url of the extensions to retrieve (if applicable)
   * @returns { Resource | undefined | Resource[] } The extensions if they exists or the extension-array
   */
  public getExtension(url?: uri): Extension[] {
    if (url) {
      return this.getExtension().filter(extension => extension.url === url)
    }
    return this.resource.extension || []
  }

  /**
   * Sets the resource's extension-array to the provided array
   * If no argument is provided, the field will be removed
   * @param { Extension[] } extension The array to be set as the resource's extension array
   * @returns the Builder
   */
  public setExtension(extension?: Extension[]): this {
    extension
      ? (this.resource.extension = extension)
      : delete this.resource.extension
    return this
  }

  /**
   * Adds the extension provided as an argument to the resource's extension array
   * @param { Extension } extension The extension to be added to the extension-array
   * @returns the Builder
   */
  public addExtension(extension: Extension): this {
    this.setExtension([...this.getExtension(), extension])
    return this
  }

  /**
   * Returns either the modifierExtensions identified by the provided url from the modifierExtension array or the entire modifierExtension array, if no url is provided
   * @param { uri } [url] The url of the modifierExtensions to retrieve (if applicable)
   * @returns { Resource | undefined | Resource[] } The modifierExtensions if they exists or the modifierExtension-array
   */
  public getModifierExtension(url?: uri): Extension[] {
    if (url) {
      return this.getModifierExtension().filter(
        modifierExtension => modifierExtension.url === url
      )
    }
    return this.resource.modifierExtension || []
  }

  /**
   * Sets the resource's modifierExtension-array to the provided array
   * If no argument is provided, the field will be removed
   * @param { Extension[] } modifierExtension The array to be set as the resource's modifierExtension array
   * @returns the Builder
   */
  public setModifierExtension(modifierExtension?: Extension[]): this {
    modifierExtension
      ? (this.resource.modifierExtension = modifierExtension)
      : delete this.resource.modifierExtension
    return this
  }

  /**
   * Adds the extension provided as an argument to the resource's modifierExtension array
   * @param { Extension } modifierExtension The modifierExtension to be added to the modifierExtension-array
   * @returns the Builder
   */
  public addModifierExtension(modifierExtension: Extension): this {
    this.setModifierExtension([
      ...this.getModifierExtension(),
      modifierExtension,
    ])
    return this
  }

  /**
   * Make a new object that implements a FHIR DomainResource
   * @returns { DomainResource } - an object that implements FHIR DomainResource
   * @override
   */
  protected make(): DomainResource {
    const domainResource = {
      resourceType: 'DomainResource',
    } as DomainResource
    return domainResource
  }

  /**
   * Builds and returns the domainResource
   * @returns { DomainResource } - the domainResource
   * @override
   */
  public build(): DomainResource {
    return this.resource
  }
}
