import { Element, id, Extension, url } from 'fhir'

/**
 * Class representing a FHIR Element builder
 */
export class ElementBuilder {
  /**
   * The Element that is being built
   * @protected
   * @type { Element }
   * @ignore
   */
  protected element: Element

  /**
   * Creates a new FHIR ElementBuilder
   */
  public constructor() {
    this.element = this.make()
  }

  /**
   * Sets a copy of an existing element for this ElementBuilder to allow for manipulating, or interacting with, the
   * element through this ElementBuilder
   * If no argument is provided, a new Element will be generated
   * @param { Element } element - The element to set
   * @returns { ElementBuilder } The builder
   */
  public use(element?: Element): this {
    this.element = element ? { ...element } : this.make()
    return this
  }

  /**
   * Returns the element's id
   * @returns { id } The element's id
   */
  public getId(): id | undefined {
    return this.element.id
  }

  /**
   * Sets the element's id to the id provided as an argument
   * If no argument is provided, the field will be removed
   * @param { id } id the id to be set
   * @returns the Builder
   */
  public setId(id?: id): this {
    id ? (this.element.id = id) : delete this.element.id
    return this
  }

  /**
   * Returns either the extensions identified by the provided url from the extension array or the entire extension array, if no url is provided
   * @returns { Resource[] } The extensions if they exists or the extension-array
   */
  public getExtension(url?: url): Extension[] {
    if (url) {
      return this.getExtension().filter(extension => extension.url === url)
    }
    return this.element.extension || []
  }

  /**
   * Sets the resource'extension-array to the provided array
   * If no argument is provided, the field will be removed
   * @param { Extension[] } extension The array to be set as the resource's extension array
   * @returns the Builder
   */
  public setExtension(extension?: Extension[]): this {
    extension
      ? (this.element.extension = extension)
      : delete this.element.extension
    return this
  }

  /**
   * Adds the extension provided as an argument to the resource's extension array
   * @param extension The resource to be added to the extension array
   * @returns the Builder
   */
  public addExtension(extension: Extension): this {
    this.setExtension([...this.getExtension(), extension])
    return this
  }

  /**
   * Returns whether or not the element provided as an argument is in conflict with the element used in the builder
   * @param { Element } element The element to test for conflict
   * @returns { boolean } Whether it is in conflict
   */
  public conflictsWith(element: Element) {
    const elementBuilder: ElementBuilder = new ElementBuilder().use(element)

    if (this.getId() && this.getId() === elementBuilder.getId()) {
      return true
    }
    return false
  }

  /**
   * Make a new object that implements a FHIR Element
   * @returns { Element } - an object that implements FHIR Element
   */
  protected make(): Element {
    return {} as Element
  }

  /**
   * Builds and returns the element
   * @returns { Element } - the element
   */
  public build(): Element {
    return this.element
  }
}
