import {
  Identifier,
  IdentifierUse,
  CodeableConcept,
  Period,
  Reference,
} from 'fhir'
import { ElementBuilder } from './element-builder'

/** Class representing a FHIR Identifier builder */
export class IdentifierBuilder extends ElementBuilder {
  /**
   * The identifier being built
   * @private
   * @type { Identifier }
   * @ignore
   */
  protected element: Identifier

  /**
   * Create a new FHIR IdentifierBuilder
   */
  public constructor() {
    super()
    this.element = this.make()
  }

  /**
   * Sets a copy of an existing identifier for this builder to allow for manipulation
   * If no argument is provided, a new Identifier will be generated
   * @param { Identifier } identifier - the identifier to be set
   * @returns { IdentifierBuilder } The builder
   * @override
   */
  public use(identifier?: Identifier): this {
    this.element = identifier ? { ...identifier } : this.make()
    return this
  }

  /**
   * Returns the use-field of the Identifier
   * @returns { IdentifierUse } The use-field
   */
  public getUse(): IdentifierUse | undefined {
    return this.element.use
  }

  /**
   * Sets the use-field of the Identifier
   * If no argument is provided, the field will be removed
   * @param { IdentifierUse } use - The value to be set in the use-field
   * @returns { IdentifierBuilder } - The builder
   */
  public setUse(use?: IdentifierUse): this {
    use ? (this.element.use = use) : delete this.element.use
    return this
  }

  /**
   * Returns the type-field of the Identifier
   * @returns { CodeableConcept } The type-field
   */
  public getType(): CodeableConcept | undefined {
    return this.element.type
  }

  /**
   * Sets the type-field of the Identifier
   * If no argument is provided, the field will be removed
   * @param { CodeableConcept } type - The value to be set in the type-field
   * @returns { IdentifierBuilder } - The builder
   */
  public setType(type?: CodeableConcept): this {
    type ? (this.element.type = type) : delete this.element.type
    return this
  }

  /**
   * Returns the system-field of the Identifier
   * @returns { string } The system-field
   */
  public getSystem(): string | undefined {
    return this.element.system
  }

  /**
   * Sets the system-field of the Identifier
   * If no argument is provided, the field will be removed
   * @param { string } system - The value to be set in the system-field
   * @returns { IdentifierBuilder } - The builder
   */
  public setSystem(system?: string): this {
    system ? (this.element.system = system) : delete this.element.system
    return this
  }

  /**
   * Returns the value-field of the Identifier
   * @returns { string } The value-field
   */
  public getValue(): string | undefined {
    return this.element.value
  }

  /**
   * Sets the value-field of the Identifier
   * If no argument is provided, the field will be removed
   * @param { string } value - The value to be set in the value-field
   * @returns { IdentifierBuilder } - The builder
   */
  public setValue(value?: string): this {
    value ? (this.element.value = value) : delete this.element.value
    return this
  }

  /**
   * Returns the period-field of the Identifier
   * @returns { Period } The period-field
   */
  public getPeriod(): Period | undefined {
    return this.element.period
  }

  /**
   * Sets the period-field of the Identifier
   * If no argument is provided, the field will be removed
   * @param { Period } period - The value to be set in the period-field
   * @returns { IdentifierBuilder } - The builder
   */
  public setPeriod(period?: Period): this {
    period ? (this.element.period = period) : delete this.element.period
    return this
  }

  /**
   * Returns the assigner-field of the Identifier
   * @returns { Reference } The assigner-field
   */
  public getAssigner(): Reference | undefined {
    return this.element.assigner
  }

  /**
   * Sets the assigner-field of the Identifier
   * If no argument is provided, the field will be removed
   * @param { Reference } assigner - The value to be set in the assigner-field
   * @returns { IdentifierBuilder } - The builder
   */
  public setAssigner(assigner?: Reference): this {
    assigner ? (this.element.assigner = assigner) : delete this.element.assigner
    return this
  }

  /**
   * Returns whether or not the identifier provided as an argument is in conflict with the identifier used in the builder
   * @param { Identifier } identifier The identifier to test for conflict
   * @returns { boolean } Whether it is in conflict
   * @override
   */
  public conflictsWith(identifier: Identifier) {
    let conflict = super.conflictsWith(identifier)
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use(
      identifier
    )

    if (
      this.getSystem() &&
      this.getValue() &&
      identifierBuilder.getSystem() &&
      identifierBuilder.getValue()
    ) {
      conflict =
        this.getSystem() === identifierBuilder.getSystem() &&
        this.getValue() === identifierBuilder.getValue()
    }
    return conflict
  }

  /**
   * Creates a new Identifier
   * @returns { Identifier } The identifier
   * @override
   */
  protected make(): Identifier {
    return {} as Identifier
  }

  /**
   * Returns the identifier
   * @returns { Identifier } the identifier
   * @override
   */
  public build(): Identifier {
    return this.element
  }
}
