import { Identifier, Reference } from 'fhir'
import { ElementBuilder } from './element-builder'
import { IdentifierBuilder } from './identifier-builder'

/** Class representing a FHIR  */
export class ReferenceBuilder extends ElementBuilder {
  /**
   * The Reference that is being built
   * @protected
   * @type { Reference }
   * @ignore
   */
  protected element: Reference

  /**
   * Create a new FHIR ReferenceBuilder
   */
  public constructor() {
    super()
    this.element = this.make()
  }

  /**
   * Sets a copy of an existing reference for this builder to allow for manipulation
   * If no argument is provided, a new Reference will be generated
   * @param { Reference } coding - the reference to be set
   * @returns { ReferenceBuilder } The builder
   * @override
   */
  public use(reference?: Reference): this {
    this.element = reference ? { ...reference } : this.make()
    return this
  }

  /**
   * Returns the reference-field of the Reference
   * @returns { string } The reference-field
   */
  public getReference(): string | undefined {
    return this.element.reference
  }

  /**
   * Sets the reference-field of the Reference
   * If no argument is provided, the field will be removed
   * @param { string } reference - The value to be set in the reference-field
   * @returns { ReferenceBuilder } - This ReferenceBuilder
   */
  public setReference(reference?: string): this {
    reference
      ? (this.element.reference = reference)
      : delete this.element.reference
    return this
  }

  /**
   * Returns the type-field of the Reference
   * @returns { string } The type-field
   */
  public getType(): string | undefined {
    return this.element.type
  }

  /**
   * Sets the type-field of the Reference
   * If no argument is provided, the field will be removed
   * @param { string } type - The value to be set in the type-field
   * @returns { ReferenceBuilder } - This ReferenceBuilder
   */
  public setType(type?: string): this {
    type ? (this.element.type = type) : delete this.element.type
    return this
  }

  /**
   * Returns the identifier-field of the Reference
   * @returns { Identifier } The identifier-field
   */
  public getIdentifier(): Identifier | undefined {
    return this.element.identifier
  }

  /**
   * Sets the identifier-field of the Reference
   * If no argument is provided, the field will be removed
   * @param { Identifier } identifier - The value to be set in the identifier-field
   * @returns { ReferenceBuilder } - This ReferenceBuilder
   */
  public setIdentifier(identifier?: Identifier): this {
    identifier
      ? (this.element.identifier = identifier)
      : delete this.element.identifier
    return this
  }

  /**
   * Returns the display-field of the Reference
   * @returns { string } The display-field
   */
  public getDisplay(): string | undefined {
    return this.element.display
  }

  /**
   * Sets the display-field of the Reference
   * If no argument is provided, the field will be removed
   * @param { string } display - The value to be set in the display-field
   * @returns { ReferenceBuilder } - This ReferenceBuilder
   */
  public setDisplay(display?: string): this {
    display ? (this.element.display = display) : delete this.element.display
    return this
  }

  /**
   * Returns whether or not the reference provided as an argument is in conflict with the reference used in the builder
   * @param { Reference } reference The reference to test for conflict
   * @returns { boolean } Whether it is in conflict
   * @override
   */
  public conflictsWith(reference: Reference) {
    let conflict = super.conflictsWith(reference)
    const referenceBuilder: ReferenceBuilder = new ReferenceBuilder().use(
      reference
    )
    if (this.getReference() && referenceBuilder.getReference()) {
      conflict = this.getReference() === referenceBuilder.getReference()
    }
    if (this.getIdentifier() && referenceBuilder.getIdentifier()) {
      conflict = new IdentifierBuilder()
        .use(this.getIdentifier())
        .conflictsWith(referenceBuilder.getIdentifier() as Identifier)
    }
    return conflict
  }

  /**
   * Make a new object that implements a FHIR Reference
   * @returns { Reference } - an object that implements FHIR Reference
   * @override
   */
  protected make(): Reference {
    const reference = {} as Reference
    return reference
  }

  /**
   * Returns the reference
   * @returns { Reference } The reference
   * @override
   */
  public build(): Reference {
    return this.element
  }
}
