import { Resource, id, code, Meta, uri } from 'fhir'

/**
 * Class representing a FHIR Resource builder
 */
export class ResourceBuilder {
  /**
   * The Resource that is being built
   * @protected
   * @type { Resource }
   * @ignore
   */
  protected resource: Resource

  /**
   * Creates a new FHIR ResourceBuilder
   */
  public constructor() {
    this.resource = this.make()
  }

  /**
   * Sets a copy of an existing resource for this builder to allow for manipulation
   * If no argument is provided, a new Resource will be generated
   * @param { Resource } resource - the resource to be set
   * @returns { ResourceBuilder } The builder
   * @override
   */
  public use(resource?: Resource): this {
    this.resource = resource ? { ...resource } : this.make()
    return this
  }

  /**
   * Returns the resourceType-field of the Resource
   * @returns { code } The resourceType-field
   */
  public getResourceType(): code | undefined {
    return this.resource.resourceType
  }

  /**
   * Sets the resourceType-field of the Resource
   * If no argument is provided, the field will be removed
   * @param { code } resourceType - The value to be set in the resourceType-field
   * @returns { ResourceBuilder } - The builder
   */
  public setResourceType(resourceType?: code): this {
    resourceType
      ? (this.resource.resourceType = resourceType)
      : delete this.resource.resourceType
    return this
  }

  /**
   * Returns the resource's id
   * @returns { id } The resource's id
   */
  public getId(): id | undefined {
    return this.resource.id
  }

  /**
   * Sets the resource's id to the id provided as an argument
   * If no argument is provided, the field will be removed
   * @param { id } id the id to be set
   * @returns the Builder
   */
  public setId(id?: id): this {
    id ? (this.resource.id = id) : delete this.resource.id
    return this
  }

  /**
   * Returns the meta-field of the Resource
   * @returns { Meta } The meta-field
   */
  public getMeta(): Meta | undefined {
    return this.resource.meta
  }

  /**
   * Sets the meta-field of the Resource
   * If no argument is provided, the field will be removed
   * @param { Meta } meta - The value to be set in the meta-field
   * @returns { ResourceBuilder } - The builder
   */
  public setMeta(meta?: Meta): this {
    meta ? (this.resource.meta = meta) : delete this.resource.meta
    return this
  }

  /**
   * Returns the implicitRules-field of the Resource
   * @returns { uri } The implicitRules-field
   */
  public getImplicitRules(): uri | undefined {
    return this.resource.implicitRules
  }

  /**
   * Sets the implicitRules-field of the Resource
   * If no argument is provided, the field will be removed
   * @param { uri } implicitRules - The value to be set in the implicitRules-field
   * @returns { ResourceBuilder } - The builder
   */
  public setImplicitRules(implicitRules?: uri): this {
    implicitRules
      ? (this.resource.implicitRules = implicitRules)
      : delete this.resource.implicitRules
    return this
  }

  /**
   * Returns the language-field of the Resource
   * @returns { code } The language-field
   */
  public getLanguage(): code | undefined {
    return this.resource.language
  }

  /**
   * Sets the language-field of the Resource
   * If no argument is provided, the field will be removed
   * @param { string } code - The value to be set in the language-field
   * @returns { ResourceBuilder } - The builder
   */
  public setLanguage(language?: code): this {
    language
      ? (this.resource.language = language)
      : delete this.resource.language
    return this
  }

  /**
   * Returns whether or not the resource provided as an argument is in conflict with the resource used in the builder
   * @param { Resource } resource The resource to test for conflict
   * @returns { boolean } Whether it is in conflict
   */
  public conflictsWith(resource: Resource) {
    const resourceBuilder: ResourceBuilder = new ResourceBuilder().use(resource)

    if (this.getId() && this.getId() === resourceBuilder.getId()) {
      return true
    }
    return false
  }

  /**
   * Make a new object that implements a FHIR Resource
   * @returns { Resource } - an object that implements FHIR Resource
   */
  protected make(): Resource {
    const resource = {
      resourceType: 'Resource',
    } as Resource
    return resource
  }

  /**
   * Builds and returns the resource
   * @returns { Resource } - the resource
   */
  public build(): Resource {
    return this.resource
  }
}
