import { Basic, dateTime, Reference, Extension, canonical } from 'fhir'
import { DomainResourceBuilder } from './domain-resource-builder'
import { _generateIsoDate } from './utils/date-util'
import { MalformedError } from './exceptions/malformed-error'
import { ExtensionBuilder } from './extension-builder'
import { Constants } from './global/constants'
import { isReference } from './global/typeguards'

/**
 * Class representing a FHIR ThresholdSet builder
 */
export class ThresholdSetBuilder extends DomainResourceBuilder {
  /**
   * The ThresholdSet that is being built
   * @protected
   * @type { Basic }
   * @ignore
   */
  protected resource: Basic

  /**
   * Creates a new FHIR ThresholdSetBuilder
   */
  public constructor() {
    super()
    this.resource = this.make()
  }

  /**
   * Sets a copy of an existing thresholdSet for this ThresholdSetBuilder to allow for manipulating, or interacting with, the
   * thresholdSet through this ThresholdSetBuilder
   * If no argument is provided, a new ThresholdSet will be generated
   * @param { ThresholdSet } thresholdSet - The thresholdSet to set
   * @returns { ThresholdSetBuilder } The builder
   * @override
   */
  public use(thresholdSet?: Basic): this {
    this.resource = thresholdSet ? { ...thresholdSet } : this.make()
    return this
  }

  /**
   * Returns the version of the ThresholdSet
   * @returns { string } The version
   */
  public getVersion(): string | undefined {
    const extension = this.getExtension(Constants.thresholdSet.version)
    if (extension.length < 1) return
    if (extension.length > 1 || !extension[0].valueString) {
      throw new MalformedError(
        'The thresholdSet contains a malformed version-extension'
      )
    }
    return extension[0].valueString
  }

  /**
   * Sets the version of the ThresholdSet
   * If no argument is provided, the field will be removed
   * @param { string } version - The version
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setVersion(system?: string): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.version
      ),
    ])
    if (system) {
      this.addExtension(
        new ExtensionBuilder()
          .setUrl(Constants.thresholdSet.version)
          .setValueString(system)
          .build()
      )
    }
    return this
  }

  /**
   * Returns the status of the ThresholdSet
   * @returns { string } The status
   */
  public getStatus(): ThresholdSetStatus | undefined {
    const extension = this.getExtension(Constants.thresholdSet.status)
    if (extension.length < 1) return
    if (extension.length > 1 || !extension[0].valueString) {
      throw new MalformedError(
        'The thresholdSet contains a malformed status-extension'
      )
    }
    const status = extension[0].valueString
    if (
      status !== 'draft' &&
      status !== 'active' &&
      status !== 'retired' &&
      status !== 'unknown'
    ) {
      throw new MalformedError(
        'The thresholdSet contains a malformed status-extension'
      )
    }
    return status
  }

  /**
   * Sets the status of the ThresholdSet
   * If no argument is provided, the field will be removed
   * @param { string } status - The version
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setStatus(status?: ThresholdSetStatus): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.status
      ),
    ])
    if (status) {
      this.addExtension(
        new ExtensionBuilder()
          .setUrl(Constants.thresholdSet.status)
          .setValueString(status)
          .build()
      )
    }
    return this
  }

  /**
   * Returns the dateTime the thresholdSet was approved
   * @returns { dateTime } The approvalDate
   */
  public getApprovalDate(): dateTime | undefined {
    const extension = this.getExtension(Constants.thresholdSet.approvalDate)
    if (extension.length < 1) return
    if (extension.length > 1 || !extension[0].valueDateTime) {
      throw new MalformedError(
        'The thresholdSet contains a malformed approvalDate-extension'
      )
    }
    return extension[0].valueDateTime
  }

  /**
   * Sets the dateTime the thresholdSet was approved
   * If no argument is provided, the field will be removed
   * @param { dateTime } approvalDate - The approvalDate
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setApprovalDate(approvalDate?: dateTime): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.approvalDate
      ),
    ])
    if (approvalDate) {
      this.addExtension(
        new ExtensionBuilder()
          .setUrl(Constants.thresholdSet.approvalDate)
          .setValueDateTime(approvalDate)
          .build()
      )
    }
    return this
  }

  /**
   * Returns a reference to the practitioner who approved the thresholdSet
   * @returns { Reference } The reference to the approving practitioner
   */
  public getApprovingPractitioner(): Reference | undefined {
    const extension = this.getExtension(
      Constants.thresholdSet.approvingPractitioner
    )
    if (extension.length < 1) return
    if (extension.length > 1 || !extension[0].valueReference) {
      throw new MalformedError(
        'The thresholdSet contains a malformed approvingPractitioner-extension'
      )
    }
    return extension[0].valueReference
  }

  /**
   * Sets a reference to the practitioner who approved the thresholdSet
   * If no argument is provided, the field will be removed
   * @param { Reference } approvingPractitioner - The reference to the approving practitioner
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setApprovingPractitioner(approvingPractitioner?: Reference): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension =>
          extension.url !== Constants.thresholdSet.approvingPractitioner
      ),
    ])
    if (approvingPractitioner) {
      this.addExtension(
        new ExtensionBuilder()
          .setUrl(Constants.thresholdSet.approvingPractitioner)
          .setValueReference(approvingPractitioner)
          .build()
      )
    }
    return this
  }

  /**
   * Returns a reference or canonical reference to the resource to which this thresholdSet applies
   * @returns { canonical | Reference } The reference to the resource to which this thresholdSet applies
   */
  public getAppliesTo(): canonical | Reference | undefined {
    const extension = this.getExtension(Constants.thresholdSet.appliesTo)
    if (extension.length < 1) return
    if (
      extension.length > 1 ||
      (!extension[0].valueReference && !extension[0].valueCanonical)
    ) {
      throw new MalformedError(
        'The thresholdSet contains a malformed appliesTo-extension'
      )
    }
    return extension[0].valueReference || extension[0].valueCanonical
  }

  /**
   * Sets a reference or canonical reference to the resource, to which this thresholdSet applies
   * If no argument is provided, the field will be removed
   * @param { canonical | Reference } appliesTo - The reference or canonical reference to the resource, to which this thresholdSet applies
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setAppliesTo(appliesTo?: canonical | Reference): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.appliesTo
      ),
    ])
    if (appliesTo) {
      const extensionBuilder = new ExtensionBuilder().setUrl(
        Constants.thresholdSet.appliesTo
      )

      if (typeof appliesTo === 'string') {
        extensionBuilder.setValueCanonical(appliesTo)
      } else if (isReference(appliesTo)) {
        extensionBuilder.setValueReference(appliesTo)
      } else {
        throw new MalformedError(
          'Provided argument is neither a canonical or a reference'
        )
      }
      this.addExtension(extensionBuilder.build())
    }
    return this
  }

  /**
   * Returns an array of thresholds in the form of extensions
   * @returns { Extension[] } The array of thresholds as extensions
   */
  public getThreshold(): Extension[] {
    return this.getExtension(Constants.thresholdSet.threshold)
  }

  /**
   * Sets the threshold array of the thresholdSet
   * If no argument is provided, the field will be removed
   * @param { Extension[] } threshold - The thresholdArray
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setThreshold(threshold?: Extension[]): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.threshold
      ),
    ])

    if (threshold) {
      this.setExtension([...this.getExtension(), ...threshold])
    }
    return this
  }

  /**
   * Adds a threshold to the threshold array of the thresholdSet
   * @param { Extension } threshold - The threshold to add to the threshold-array
   * @returns { ThresholdSetBuilder } - The builder
   */
  public addThreshold(threshold: Extension): this {
    this.setThreshold([...this.getThreshold(), threshold])

    return this
  }

  /**
   * Returns whether or not the thresholdSet provided as an argument is in conflict with the thresholdSet used in the builder
   * @param { Basic } thresholdSet The thresholdSet to test for conflict
   * @returns { boolean } Whether it is in conflict
   * @override
   */
  public conflictsWith(thresholdSet: Basic) {
    let conflict = super.conflictsWith(thresholdSet)
    const extension =
      this.getExtension(Constants.thresholdSet.appliesTo)[0] || undefined
    const otherExtension =
      new ThresholdSetBuilder()
        .use(thresholdSet)
        .getExtension(Constants.thresholdSet.appliesTo)[0] || undefined

    if (
      extension &&
      otherExtension &&
      extension.valueCanonical &&
      otherExtension.valueCanonical &&
      extension.valueCanonical === otherExtension.valueCanonical
    ) {
      conflict = true
    }

    return conflict
  }

  /**
   * Creates an ISO-string for current time
   */
  public _generateIsoDate(): string {
    return _generateIsoDate()
  }

  /**
   * Make a new object that implements a FHIR ThresholdSet
   * @returns { ThresholdSet } - an object that implements FHIR ThresholdSet
   * @override
   */
  protected make(): Basic {
    return {
      resourceType: 'Basic',
      created: this._generateIsoDate(),
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    } as Basic
  }

  /**
   * Builds and returns the plandefinition
   * @returns { ThresholdSet } - the plandefinition
   * @override
   */
  public build(): Basic {
    if (this.getExtension().length === 0) this.setExtension()
    return this.resource
  }
}

/**
 * All possible values of ThresholdSet.status
 */
type ThresholdSetStatus = 'draft' | 'active' | 'retired' | 'unknown'
