import { Extension, CodeableConcept } from 'fhir'
import { _generateIsoDate } from './utils/date-util'
import { MalformedError } from './exceptions/malformed-error'
import { ExtensionBuilder } from './extension-builder'
import { ElementBuilder } from './element-builder'
import { Constants } from './global/constants'

/**
 * Class representing a Threshold builder for Fhir ThresholdSets
 */
export class ThresholdSetThresholdBuilder extends ElementBuilder {
  /**
   * The threshold that is being built
   * @protected
   * @type { Extension }
   * @ignore
   */
  protected element: Extension

  /**
   * Creates a new Threshold for FHIR ThresholdSets
   */
  public constructor() {
    super()
    this.element = this.make()
  }

  /**
   * Sets a copy of an existing threshold for this builder to allow for manipulating, or interacting with, the
   * threshold through this builder
   * If no argument is provided, a new ThresholdSetThreshold will be generated
   * @param { Extension } threshold - The threshold to set
   * @returns { ThresholdSetBuilder } The builder
   * @override
   */
  public use(threshold?: Extension): this {
    this.element = threshold ? { ...threshold } : this.make()
    return this
  }

  /**
   * Returns the type of the threshold
   * @returns { CodeableConcept } The version
   */
  public getType(): CodeableConcept | undefined {
    const extension = this.getExtension(Constants.thresholdSet.type)
    if (extension.length < 1) return
    if (extension.length > 1 || !extension[0].valueCodeableConcept) {
      throw new MalformedError(
        'The threshold contains a malformed type-extension'
      )
    }
    return extension[0].valueCodeableConcept
  }

  /**
   * Sets the type of the threshold
   * If no argument is provided, the field will be removed
   * @param { CodeableConcept } type - The type
   * @returns { ThresholdSetBuilder } - The builder
   */
  public setType(type?: CodeableConcept): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.type
      ),
    ])
    if (type) {
      this.addExtension(
        new ExtensionBuilder()
          .setUrl(Constants.thresholdSet.type)
          .setValueCodeableConcept(type)
          .build()
      )
    }
    return this
  }

  /**
   * Returns an array of effectiveWhens in the form of extensions
   * @returns { Extension[] } The array of effectiveWhens as extensions
   */
  public getEffectiveWhen(): Extension[] {
    return this.getExtension(Constants.thresholdSet.effectiveWhen)
  }

  /**
   * Sets the effectiveWhen array of the thresholdSet
   * If no argument is provided, the field will be removed
   * @param { Extension[] } threshold - The effectiveWhen array
   * @returns { ThresholdSetThresholdBuilder } - The builder
   */
  public setEffectiveWhen(effectiveWhen?: Extension[]): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.effectiveWhen
      ),
    ])

    if (effectiveWhen) {
      this.setExtension([...this.getExtension(), ...effectiveWhen])
    }
    return this
  }

  /**
   * Adds an effectiveWhen to the effectiveWhen array of the threshold
   * @param { Extension } effectiveWhen - The effectiveWhen to add to the effectiveWhen-array
   * @returns { ThresholdSetThresholdBuilder } - The builder
   */
  public addEffectiveWhen(effectiveWhen: Extension): this {
    this.setEffectiveWhen([...this.getEffectiveWhen(), effectiveWhen])

    return this
  }

  /**
   * Returns whether the threshold is in effect if all or any of the effectiveWhens are in effect
   * @returns { ThresholdEffectiveBehavior } The effectiveBehavior
   */
  public getEffectiveBehavior(): ThresholdEffectiveBehavior | undefined {
    const extension = this.getExtension(
      Constants.thresholdSet.effectiveBehavior
    )
    if (extension.length < 1) return
    if (extension.length > 1 || !extension[0].valueCode) {
      throw new MalformedError(
        'The threshold contains a malformed type-extension'
      )
    }
    const effectiveBehavior = extension[0].valueCode
    if (effectiveBehavior !== 'all' && effectiveBehavior !== 'any') {
      throw new MalformedError(
        'The threshold contains a malformed effectiveBehavior-extension'
      )
    }
    return effectiveBehavior
  }

  /**
   * Sets whether the threshold is in effect if all or any of the effectiveWhens are in effect
   * If no argument is provided, the field will be removed
   * @param { ThresholdEffectiveBehavior } effectiveBehavior - The effectiveBehavior
   * @returns { ThresholdSetThresholdBuilder } - The builder
   */
  public setEffectiveBehavior(
    effectiveBehavior?: ThresholdEffectiveBehavior
  ): this {
    this.setExtension([
      ...this.getExtension().filter(
        extension => extension.url !== Constants.thresholdSet.effectiveBehavior
      ),
    ])
    if (effectiveBehavior) {
      this.addExtension(
        new ExtensionBuilder()
          .setUrl(Constants.thresholdSet.effectiveBehavior)
          .setValueCode(effectiveBehavior)
          .build()
      )
    }
    return this
  }

  /**
   * Make a new object that implements a FHIR ThresholdSet
   * @returns { ThresholdSet } - an object that implements FHIR ThresholdSet
   * @override
   */
  protected make(): Extension {
    return {
      url: Constants.thresholdSet.threshold,
      extension: [],
    }
  }

  /**
   * Builds and returns the plandefinition
   * @returns { Extension } - the plandefinition
   * @override
   */
  public build(): Extension {
    return this.element
  }
}

/**
 * All possible values of Threshold.effectiveBehavior
 */
type ThresholdEffectiveBehavior = 'all' | 'any'
