import {
  CarePlan,
  PlanDefinition,
  canonical,
  Task,
  Basic,
  Resource,
  CarePlanActivity,
  PlanDefinitionAction,
  TaskInput,
  Coding,
  CodeableConcept,
} from 'fhir'
import { MissingPropertyError, AlreadyExistsError } from '../exceptions'
import {
  CarePlanBuilder,
  PlanDefinitionBuilder,
  PlanDefinitionActionBuilder,
  TaskBuilder,
  TaskInputBuilder,
  CodeableConceptBuilder,
  CodingBuilder,
  ResourceBuilder,
  CarePlanActivityBuilder,
  ReferenceBuilder,
  ExtensionBuilder,
  DomainResourceBuilder,
} from '..'
import { Constants } from '../global/constants'

/**
 * Adds all content from the PlanDefinition to the CarePlan
 * @param { PlanDefinition } planDefinition The PlanDefinition to copy content from
 * @param { CarePlan } carePlan The CarePlan to copy content to
 * @return { CarePlan } A new CarePlan containing the content the provided CarePlan contained,
 * along with the newly copied content
 * @throws { MissingPropertyError } When PlanDefinition or CarePLan is missing necessary properties
 * @throws { Error } References malformed
 */
const planDefinitionToCarePlan: (
  pd: PlanDefinition,
  cp: CarePlan
) => CarePlan = (pd: PlanDefinition, cp: CarePlan) => {
  const pdBuilder: PlanDefinitionBuilder = new PlanDefinitionBuilder().use({
    ...pd,
  })
  const cpBuilder: CarePlanBuilder = new CarePlanBuilder().use({ ...cp })
  const errors: AlreadyExistsError[] = []

  // Create canonical reference to plandefinition
  if (!pdBuilder.getUrl() || !pdBuilder.getVersion()) {
    throw new MissingPropertyError(
      'PlanDefinition is missing either url or version-field. Cannot create canonical reference.'
    )
  }
  const canonical = `${pd.url}|${pd.version}`

  // Check for conflicts with instantiatesCanonical
  if (
    cpBuilder
      .getInstantiatesCanonical()
      .find(existing => existing === canonical)
  ) {
    throw new AlreadyExistsError(
      'CarePlan already instantiates this PlanDefinition'
    )
  }

  // For each Action
  pdBuilder.getAction().forEach(action => {
    const pdActionBuilder: PlanDefinitionActionBuilder = new PlanDefinitionActionBuilder().use(
      action
    )
    // Get ThresholdSet from PlanDefinition Contained
    const thresholdSetRef = _getThresholdSetReference(pdActionBuilder)
    const thresholdSet = thresholdSetRef
      ? pdBuilder.getContained(thresholdSetRef)
      : undefined
    // Handle conflicts between ThresholdSet and CarePlan contained and add ThresholdSet to contained
    if (!thresholdSet) {
      throw new MissingPropertyError(
        'No ThresholdSet found for PlanDefinitionAction'
      )
    }

    // Create activity from action and add it to carePlan
    try {
      cpBuilder.use(
        actionToActivity(action, thresholdSet as Basic, cpBuilder.build())
      )
    } catch (e) {
      if (e instanceof AlreadyExistsError) {
        errors.push(e)
      } else {
        throw e
      }
    }
  })

  // If errors occurred, return a new AlreadyExistsError, containing the errors that occurred
  if (errors.length > 0) {
    throw new AlreadyExistsError(
      'The content of PlanDefinition could not be added to CarePlan',
      ...errors
    )
  }
  // Add instantiatesCanonical reference
  cpBuilder.addInstantiatesCanonical(canonical)
  return cpBuilder.build()
}

/**
 * Creates an Activity from a PlanDefinitionAction and a ThresholdSet and adds it to the CarePlan
 * @param { PlanDefinitionAction }planDefinitionAction The Action from which to create the Activity
 * @param { Basic } thresholdSet The ThresholdSet for the resource
 * @param { CarePlan } carePlan The CarePlan to which the activity should be added
 * @returns { CarePlan } A new CarePlan containing the newly added activity
 * @throws { MissingPropertyError } if PlanDefinitionAction contains no definitionCanonical
 * @throws { AlreadyExistsError } if the action or thresholdSet is in conflict with the existing resources on the CarePlan
 */
const actionToActivity: (
  action: PlanDefinitionAction,
  thresholdSet: Basic,
  carePlan: CarePlan
) => CarePlan = (
  action: PlanDefinitionAction,
  thresholdSet: Basic,
  carePlan: CarePlan
) => {
  const pdActionBuilder: PlanDefinitionActionBuilder = new PlanDefinitionActionBuilder().use(
    action
  )
  const cpBuilder = new CarePlanBuilder().use({ ...carePlan })

  // Get canonical from action
  const canonical:
    | canonical
    | undefined = pdActionBuilder.getDefinitionCanonical()
  if (!canonical) {
    throw new MissingPropertyError(
      'PlanDefinitionAction contains no definitionCanonical'
    )
  }

  // Create Task
  let task: Task = _createTask(canonical)

  // Add Task to Contained-array
  task = _handleIdConflicts(task, cpBuilder.build()) as Task
  cpBuilder.addContained(task)

  // Add ThresholdSet to Contained-array
  thresholdSet = _handleIdConflicts(thresholdSet, cpBuilder.build()) as Basic
  cpBuilder.addContained(thresholdSet)

  // Create activity with ref to Task and ThresholdSet
  const activity: CarePlanActivity = _createActivity(task.id, thresholdSet.id)
  // Add Activity to CP
  cpBuilder.addActivity(activity)
  return cpBuilder.build()
}

/**
 * Creates a task with the canonical reference provided
 * @returns { Task } The created task
 */
const _createTask: (reference: canonical) => Task = (reference: canonical) => {

  const coding = new CodingBuilder()
    .setSystem('http://hl7.org/fhir/resource-types')
    .setCode('Questionnaire')
    .setDisplay('Questionnaire')
    .build()

  const codeableConcept = new CodeableConceptBuilder()
    .addCoding(coding)
    .setText('Questionnaire')
    .build()

  const taskInput = new TaskInputBuilder()
    .setType(codeableConcept)
    .setValueCanonical(reference)
    .build()

  return new TaskBuilder()
    .setStatus('requested')
    .setIntent('order')
    .setAuthoredOn(new Date().toISOString())
    .setId(`task-${_generateRandomFourCharString()}`)
    .setLastModified(new Date().toISOString())
    .addInput(taskInput)
    .build()
}

/**
 * Handles potential conflicts that could arise when adding a resource to a CarePlan's contained-array
 * @param { Resource } resource The resource that is to be added to the CarePlan's contained-array
 * @param { CarePlan } carePlan The CarePlan
 * @returns The resource in a form where it can be added to the CarePlan's contained-array
 * @throws { AlreadyExistsError } If the CarePlan's contained-array already contains a resource that
 * references the same resource that the provided resource does
 */
const _handleIdConflicts: (
  resource: Resource,
  carePlan: CarePlan
) => Resource = (resource: Resource, carePlan: CarePlan) => {
  const resourceBuilder = new ResourceBuilder().use(resource)
  // recursively attempt to add contained reference
  let retry = false
  do {
    try {
      new DomainResourceBuilder()
        .use({ ...carePlan })
        .addContained(resourceBuilder.build())
      retry = false
    } catch (e) {
      retry = true
      resourceBuilder.setId(_generateRandomFourCharString())
    }
  } while (retry)
  return resourceBuilder.build()
}

/**
 * Retrieves reference to thresholdSet from PlanDefinitionAction
 * @param { PlanDefinitionActionBuilder } planDefinitionActionBuilder The planDefinitionActionBuilder
 * @returns { string } The reference
 */
const _getThresholdSetReference: (
  planDefinitionActionBuilder: PlanDefinitionActionBuilder
) => string | undefined = (
  planDefinitionActionBuilder: PlanDefinitionActionBuilder
) => {
  const extensions = planDefinitionActionBuilder.getExtension(
    Constants.planDefinition.thresholdSet
  )

  const extension = extensions[0] || undefined
  if (!extension || !extension.valueReference) return
  return extension.valueReference.reference
}

/**
 * Creates a CarePlanActivity from provided taskId and thresholdSetId
 * @param { string } taskId The id for the task
 * @param { string } thresholdSetId The id for the thresholdSet
 * @returns { CarePlanActivity } The CarePlanActivity
 */
const _createActivity: (
  taskId: string | undefined,
  thresholdSetId: string | undefined
) => CarePlanActivity = (
  taskId: string | undefined,
  thresholdSetId: string | undefined
) => {
  if (!taskId) throw new MissingPropertyError('No Taskid provided')
  if (!thresholdSetId) {
    throw new MissingPropertyError('No ThresholdSetId provided')
  }
  const taskReference = new ReferenceBuilder()
    .setReference(`#${taskId}`)
    .setType('Task')
    .build()

  const thresholdSetReference = new ReferenceBuilder()
    .setReference(`#${thresholdSetId}`)
    .build()

  const extension = new ExtensionBuilder()
    .setUrl(Constants.carePlan.thresholdSet)
    .setValueReference(thresholdSetReference)
    .build()

  return new CarePlanActivityBuilder()
    .setReference(taskReference)
    .addExtension(extension)
    .build()
}

/**
 * Generates a random four character string
 * @returns { string } A random four character string
 */
const _generateRandomFourCharString: () => string = () => {
  return Math.random()
    .toString(36)
    .substring(0, 6)
    .replace('0.', '')
}

export {
  planDefinitionToCarePlan,
  actionToActivity,
  _createTask,
  _handleIdConflicts,
  _getThresholdSetReference,
  _createActivity,
  _generateRandomFourCharString,
}
