import { expect } from 'chai'
import { Extension, BackboneElement } from 'fhir'
import { BackboneElementBuilder, ExtensionBuilder } from '../src'

describe('BackboneElementBuilder', () => {
  it('should be able to build minimal BackboneElement', () => {
    const backboneElement: BackboneElement = new BackboneElementBuilder().build()
    expect(backboneElement).to.eql({})
  })

  it('should be able to accept existing BackboneElement', () => {
    const backboneElement: BackboneElement = {
      modifierExtension: [{ url: 'http://example.com/extension' }],
    }

    expect(new BackboneElementBuilder().use(backboneElement).build()).to.eql({
      modifierExtension: [{ url: 'http://example.com/extension' }],
    })
  })

  it('should be able to reset', () => {
    const backboneElement: BackboneElement = {
      modifierExtension: [{ url: 'http://example.com/extension' }],
    }

    const backboneElementBuilder: BackboneElementBuilder = new BackboneElementBuilder().use(
      backboneElement
    )
    expect(backboneElementBuilder.use().build()).to.eql({})
  })

  it('should be able to set modifierExtension-array', () => {
    const extension: Extension = new ExtensionBuilder()
      .setUrl('http://example.com/extension')
      .build()

    const backboneElement: BackboneElement = new BackboneElementBuilder()
      .setModifierExtension([extension])
      .build()

    expect(backboneElement).to.eql({
      modifierExtension: [{ url: 'http://example.com/extension' }],
    })
  })

  it('should be able to unset modifierExtension-array', () => {
    const backboneElementBuilder: BackboneElementBuilder = new BackboneElementBuilder().use(
      {
        modifierExtension: [{ url: 'http://example.com/extension' }],
      }
    )

    expect(backboneElementBuilder.setModifierExtension().build()).to.eql({})
  })

  it('should be able to return modifierExtension-array', () => {
    const backboneElement: BackboneElement = {
      modifierExtension: [{ url: 'http://example.com/extension' }],
    }

    expect(
      new BackboneElementBuilder().use(backboneElement).getModifierExtension()
    ).to.eql([{ url: 'http://example.com/extension' }])
  })

  it('should return empty array when modifierExtension-array is empty', () => {
    expect(new BackboneElementBuilder().getModifierExtension()).to.eql([])
  })

  it('should be able to return extensions with specific url from modifierExtension-array', () => {
    const backboneElement: BackboneElement = {
      modifierExtension: [
        { url: 'http://example.com/extension' },
        { url: 'http://example.com/another-extension' },
      ],
    }

    expect(
      new BackboneElementBuilder()
        .use(backboneElement)
        .getModifierExtension('http://example.com/extension')
    ).to.eql([{ url: 'http://example.com/extension' }])
  })

  it('should be able to add extension to empty modifierExtension-array', () => {
    expect(
      new BackboneElementBuilder()
        .addModifierExtension({ url: 'http://example.com/extension' })
        .getModifierExtension()
    ).to.eql([{ url: 'http://example.com/extension' }])
  })
})
