import { expect } from 'chai'
import {
  CarePlan,
  Identifier,
  Reference,
  canonical,
  dateTime,
  CarePlanStatus,
  CarePlanIntent,
  CarePlanActivity,
  Task,
  Basic,
} from 'fhir'
import { CarePlanBuilder } from '../src/care-plan-builder'
import { CarePlanActivityBuilder } from '../src/care-plan-activity-builder'
import { ExtensionBuilder } from '../src/extension-builder'
import { ReferenceBuilder } from '../src/reference-builder'
import { AlreadyExistsError, MissingPropertyError } from '../src/exceptions'
import { TaskBuilder } from '../src/task-builder'
import { ThresholdSetBuilder } from '../src/threshold-set-builder'
import { deepcopy } from './utils'
import carePlan from './resources/CarePlan.json'
import sinon from 'sinon'
import { Constants } from 'global/constants'

describe('CarePlanBuilder', () => {
  let testCarePlan: CarePlan | undefined

  before(() => {
    // Mock make to avoid testing generated timestamp
    sinon
      .stub(CarePlanBuilder.prototype, '_generateIsoDate')
      .callsFake(() => '0000-00-00T00:00:00.000Z')
  })

  beforeEach(() => {
    testCarePlan = deepcopy(carePlan) as CarePlan
  })

  /** Building carePlans */
  it('should be able to build minimal carePlan', () => {
    const carePlan: CarePlan = new CarePlanBuilder().build()

    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  /** Setting carePlan */
  it('should be able to set carePlan', () => {
    const carePlan: CarePlan = new CarePlanBuilder().use(testCarePlan).build()
    expect(carePlan).to.eql(testCarePlan)
  })

  it('should be able to unset carePlan', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use(
      testCarePlan
    )
    const carePlan: CarePlan = carePlanBuilder.use().build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  /** identifier */

  it('should be able to set identifiers for a carePlan', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setIdentifier([
        {
          system: 'system',
          value: 'value',
        },
      ])
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      identifier: [
        {
          system: 'system',
          value: 'value',
        },
      ],
    })
  })

  it('should be able to unset identifiers for a carePlan', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      identifier: [
        {
          system: 'system',
          value: 'value',
        },
      ],
    } as CarePlan)
    expect(carePlanBuilder.setIdentifier().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to add identifier to a carePlan with no identifiers', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .addIdentifier({
        system: 'system',
        value: 'value',
      })
      .build()
    expect((carePlan.identifier as Identifier[]).length).to.eql(1)
  })

  it('should be able to add identifier to a carePlan with existing identifiers', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().addIdentifier(
      {
        system: 'existing',
        value: 'existing',
      }
    )
    carePlanBuilder.addIdentifier({ system: 'system', value: 'value' })
    expect((carePlan.identifier as Identifier[]).length).to.eql(2)
  })

  it('should throw an error when adding an identifier in a system in which there is already an identifier', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().addIdentifier(
      {
        system: 'existing',
        value: 'existing',
      }
    )
    expect(() => {
      carePlanBuilder.addIdentifier({
        system: 'existing',
        value: 'existing',
      })
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error when adding an identifier system or value', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder()
    expect(() => {
      carePlanBuilder.addIdentifier({
        value: 'existing',
      })
    }).to.throw(MissingPropertyError)
  })

  it('should return empty array when getting identifiers and none defined', () => {
    const identifiers: Identifier[] = new CarePlanBuilder().getIdentifier()
    expect(identifiers).to.eql([])
  })

  it('should be able to return array of identifiers', () => {
    const identifiers: Identifier[] = new CarePlanBuilder()
      .setIdentifier([
        {
          system: 'system',
          value: 'value',
        },
      ])
      .getIdentifier()
    expect(identifiers).to.eql([
      {
        system: 'system',
        value: 'value',
      },
    ])
  })

  it("should return undefined when getting identifier in system that isn't there", () => {
    const identifier:
      | Identifier
      | undefined = new CarePlanBuilder().getIdentifier('none existent system')
    expect(identifier).to.eql(undefined)
  })

  it('should be able return an identifier based on system', () => {
    const identifier: Identifier | undefined = new CarePlanBuilder()
      .setIdentifier([
        {
          system: 'system',
          value: 'value',
        },
      ])
      .getIdentifier('system')
    expect(identifier).to.eql({ system: 'system', value: 'value' })
  })

  /** instantiatesCanonical */

  it('should be able to set instantiatesCanonical for a carePlan', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setInstantiatesCanonical([
        'canonical://uuid/CarePlan/fcbcfe90-2014-40ad-96c3-681715d76726|1.0.0',
      ])
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      instantiatesCanonical: [
        'canonical://uuid/CarePlan/fcbcfe90-2014-40ad-96c3-681715d76726|1.0.0',
      ],
    })
  })

  it('should be able to unset instantiatesCanonical for a carePlan', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      instantiatesCanonical: [
        'canonical://uuid/CarePlan/fcbcfe90-2014-40ad-96c3-681715d76726|1.0.0',
      ],
    } as CarePlan)
    expect(carePlanBuilder.setInstantiatesCanonical().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to set add identifier to a carePlan with no identifiers', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .addInstantiatesCanonical(
        'canonical://uuid/CarePlan/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0'
      )
      .build()
    expect((carePlan.instantiatesCanonical as canonical[]).length).to.eql(1)
  })

  it('should be able to add identifier to a carePlan with existing identifiers', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().addInstantiatesCanonical(
      'canonical://uuid/CarePlan/fcbcfe90-2014-40ad-96c3-681715d76726|1.0.0'
    )
    carePlanBuilder.addInstantiatesCanonical(
      'canonical://uuid/CarePlan/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0'
    )
    expect((carePlan.identifier as Identifier[]).length).to.eql(2)
  })

  it('should throw an error when adding a canonical reference that is already there', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().addInstantiatesCanonical(
      'canonical://uuid/CarePlan/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0'
    )

    expect(() => {
      carePlanBuilder.addInstantiatesCanonical(
        'canonical://uuid/CarePlan/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0'
      )
    }).to.throw(AlreadyExistsError)
  })

  it('should return empty array when getting instantiateCanonical and none defined', () => {
    const canonicals: canonical[] = new CarePlanBuilder().getInstantiatesCanonical()
    expect(canonicals).to.eql([])
  })

  it('should be able to return array of identifiers', () => {
    const canonicals: canonical[] = new CarePlanBuilder()
      .setInstantiatesCanonical([
        'canonical://uuid/CarePlan/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0',
      ])
      .getInstantiatesCanonical()
    expect(canonicals).to.eql([
      'canonical://uuid/CarePlan/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0',
    ])
  })

  /** status */

  it('should be able to set status to one of the possible values', () => {
    const carePlan: CarePlan = new CarePlanBuilder().setStatus('active').build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      status: 'active',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return carePlan status', () => {
    const status: CarePlanStatus = new CarePlanBuilder()
      .use({
        resourceType: 'CarePlan',
        status: 'active',
        intent: 'order',
        created: '0000-00-00T00:00:00.000Z',
      } as CarePlan)
      .getStatus()
    expect(status).to.eql('active')
  })

  /** intent */

  it('should be able to set carePlan intent', () => {
    expect(new CarePlanBuilder().setIntent('plan').build()).to.eql({
      resourceType: 'CarePlan',
      intent: 'plan',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return carePlan intent', () => {
    const intent: CarePlanIntent = new CarePlanBuilder()
    .setIntent('order')
    .getIntent()
    expect(intent).to.eql('order')
  })

  /** title */

  it('should be able to set title', () => {
    const carePlan: CarePlan = new CarePlanBuilder().setTitle('title').build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      title: 'title',
    })
  })

  it('should be able to unset title', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      title: 'title',
    } as CarePlan)
    expect(carePlanBuilder.setTitle().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return carePlan title', () => {
    expect(
      new CarePlanBuilder()
        .use({
          resourceType: 'CarePlan',
          status: 'draft',
          intent: 'order',
          created: '0000-00-00T00:00:00.000Z',
          title: 'title',
        } as CarePlan)
        .getTitle()
    ).to.eql('title')
  })

  /** subject */

  it('should be able to set subject', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setSubject({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      subject: {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    })
  })

  it('should be able to unset subject', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      subject: {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    } as CarePlan)
    expect(carePlanBuilder.setSubject().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return subject of carePlan', () => {
    const subject: Reference = new CarePlanBuilder()
      .setSubject({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .getSubject()
    expect(subject).to.eql({
      identifier: {
        system: 'system',
        value: 'value',
      },
    })
  })

  /** period */

  it('should be able to set period', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setPeriod({
        start: 'startDate',
        end: 'endDate',
      })
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      period: {
        start: 'startDate',
        end: 'endDate',
      },
    })
  })

  it('should be able to unset period', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      period: {
        start: 'startDate',
        end: 'endDate',
      },
    } as CarePlan)
    expect(carePlanBuilder.setPeriod().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return period of carePlan', () => {
    expect(
      new CarePlanBuilder()
        .use({
          resourceType: 'CarePlan',
          status: 'draft',
          intent: 'order',
          created: '0000-00-00T00:00:00.000Z',
          title: 'title',
          period: {
            start: 'startDate',
            end: 'endDate',
          },
        } as CarePlan)
        .getPeriod()
    ).to.eql({
      start: 'startDate',
      end: 'endDate',
    })
  })

  /** encounter */

  it('should be able to set encouter', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setEncounter({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      encounter: {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    })
  })

  it('should be able to unset encounter', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      encounter: {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    } as CarePlan)
    expect(carePlanBuilder.setEncounter().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return encounter of carePlan', () => {
    const encounter: Reference = new CarePlanBuilder()
      .setEncounter({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .getEncounter()
    expect(encounter).to.eql({
      identifier: {
        system: 'system',
        value: 'value',
      },
    })
  })

  /** created */
  it('should be able to set created', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setCreated('2017-11-27T13:05:00.000Z')
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '2017-11-27T13:05:00.000Z',
    })
  })

  it('should be able to unset created', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '2017-11-27T13:05:00.000Z',
    } as CarePlan)
    expect(carePlanBuilder.setCreated().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
    })
  })

  it('should be able to return created', () => {
    const created: dateTime = new CarePlanBuilder()
      .setCreated('2017-11-27T13:05:00.000Z')
      .getCreated()
    expect(created).to.eql('2017-11-27T13:05:00.000Z')
  })

  /** author */

  it('should be able to set author', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setAuthor({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      author: {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    })
  })

  it('should be able to unset author', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      author: {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    } as CarePlan)
    expect(carePlanBuilder.setAuthor().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return author of carePlan', () => {
    const author: Reference = new CarePlanBuilder()
      .setAuthor({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .getAuthor()
    expect(author).to.eql({
      identifier: {
        system: 'system',
        value: 'value',
      },
    })
  })

  /** addresses */

  it('should be able to set addresses-field for a carePlan', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .setAddresses([
        {
          identifier: {
            system: 'system',
            value: 'value',
          },
        },
      ])
      .build()
    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      addresses: [
        {
          identifier: {
            system: 'system',
            value: 'value',
          },
        },
      ],
    })
  })

  it('should be able to unset addresses-field for carePlan', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      addresses: [
        {
          identifier: {
            system: 'system',
            value: 'value',
          },
        },
      ],
    } as CarePlan)
    expect(carePlanBuilder.setAddresses().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should throw an error when setting addresses-array containing an item without necessary information', () => {
    expect(() => {
      new CarePlanBuilder().setAddresses([
        {
          identifier: {
            system: 'system',
          },
        },
      ])
    }).to.throw(MissingPropertyError)
  })

  it('should be able to add an address to a carePlan with no addresses', () => {
    const carePlan: CarePlan = new CarePlanBuilder()
      .addAddresses({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .build()
    expect((carePlan.addresses as Reference[]).length).to.eql(1)
  })

  it('should be able to add an address to a carePlan with existing addresses', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().addAddresses(
      {
        identifier: {
          system: 'existing',
          value: 'existing',
        },
      }
    )
    const carePlan: CarePlan = carePlanBuilder
      .addAddresses({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
      .build()
    expect((carePlan.addresses as Reference[]).length).to.eql(2)
  })

  it('should throw an error when adding an item with a reference that already exists in the addresses-array', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      addresses: [
        {
          reference: 'http://example.com/CarePlan/7',
        },
      ],
    } as CarePlan)
    expect(() => {
      carePlanBuilder.addAddresses({
        reference: 'http://example.com/CarePlan/7',
      })
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error when adding an item with an identifier that already exists in the addresses-array', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      addresses: [
        {
          identifier: {
            system: 'system',
            value: 'value',
          },
        },
      ],
    } as CarePlan)
    expect(() => {
      carePlanBuilder
        .addAddresses({
          identifier: {
            system: 'system',
            value: 'value',
          },
        })
        .build()
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error when adding an item without necessary information to the addresses-array', () => {
    expect(() => {
      new CarePlanBuilder().addAddresses({
        identifier: {
          system: 'system',
        },
      })
    }).to.throw(MissingPropertyError)
  })

  it('should throw an error when adding an address with an identifier with only either system or value', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder()
    expect(() => {
      carePlanBuilder.addAddresses({
        identifier: {
          value: 'value',
        },
      })
    }).to.throw(MissingPropertyError)
    expect(() => {
      carePlanBuilder.addAddresses({
        identifier: {
          system: 'system',
        },
      })
    }).to.throw(MissingPropertyError)
  })

  it('should return empty array when getting addresses and none defined', () => {
    const addresses: Reference[] = new CarePlanBuilder().getAddresses()
    expect(addresses).to.eql([])
  })

  it('should be able to return an array of addresses', () => {
    const addresses: Reference[] = new CarePlanBuilder()
      .setAddresses([
        {
          identifier: {
            system: 'system',
            value: 'value',
          },
        },
      ])
      .getAddresses()
    expect(addresses).to.eql([
      {
        identifier: {
          system: 'system',
          value: 'value',
        },
      },
    ])
  })

  it('should be able to set activity-array', () => {
    const activity: CarePlanActivity = new CarePlanActivityBuilder()
      .setId('id')
      .build()

    const carePlan: CarePlan = new CarePlanBuilder()
      .setActivity([activity])
      .build()

    expect(carePlan).to.eql({
      resourceType: 'CarePlan',
      created: '0000-00-00T00:00:00.000Z',
      activity: [{ id: 'id' }],
    })
  })

  it('should be able to unset activity-array', () => {
    const carePlanBuilder: CarePlanBuilder = new CarePlanBuilder().use({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      activity: [{ id: 'id' }],
    } as CarePlan)

    expect(carePlanBuilder.setActivity().build()).to.eql({
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
    })
  })

  it('should be able to return activity-array', () => {
    const carePlan: CarePlan = {
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      activity: [{ id: 'id' }],
    } as CarePlan

    expect(new CarePlanBuilder().use(carePlan).getActivity()).to.eql([
      { id: 'id' },
    ])
  })

  it('should return empty array when activity-array is empty', () => {
    expect(new CarePlanBuilder().getActivity()).to.eql([])
  })

  it('should be able to return activity with specific id from activity-array', () => {
    const carePlan: CarePlan = {
      resourceType: 'CarePlan',
      status: 'draft',
      intent: 'order',
      created: '0000-00-00T00:00:00.000Z',
      activity: [{ id: 'id' }, { id: 'id2' }],
    } as CarePlan

    expect(new CarePlanBuilder().use(carePlan).getActivity('id')).to.eql({
      id: 'id',
    })
  })

  it('should be able to add activity to empty activity-array', () => {
    expect(
      new CarePlanBuilder().addActivity({ id: 'id' }).getActivity()
    ).to.eql([{ id: 'id' }])
  })

  it('should throw an error when adding activity that is in conflict with existing activity', () => {
    expect(() => {
      new CarePlanBuilder()
        .setActivity([
          new CarePlanActivityBuilder()
            .addExtension(
              new ExtensionBuilder()
                .setUrl(Constants.carePlan.thresholdSet)
                .setValueReference(
                  new ReferenceBuilder().setReference('#ref').build()
                )
                .build()
            )
            .build(),
        ])
        .addActivity(
          new CarePlanActivityBuilder()
            .addExtension(
              new ExtensionBuilder()
                .setUrl(Constants.carePlan.thresholdSet)
                .setValueReference(
                  new ReferenceBuilder().setReference('#ref').build()
                )
                .build()
            )
            .build()
        )
    }).to.throw(AlreadyExistsError)
  })

  it('should allow adding containedResources that are not in conflict', () => {
    const cpBuilder = new CarePlanBuilder().use(testCarePlan)
    const thresholdSet = new ThresholdSetBuilder().build()
    cpBuilder.addContained(thresholdSet)
    expect(
      cpBuilder.getContained().filter(resource => resource === thresholdSet)
        .length
    ).to.eql(1)
  })

  it('should throw an error when adding a thresholdSet that is in conflict with an existing thresholdSet', () => {
    const cpBuilder = new CarePlanBuilder().use(testCarePlan)
    const thresholdSet = new ThresholdSetBuilder()
      .use(cpBuilder.getContained('thresholdSetb') as Basic)
      .setId('thresholdSet2')
      .build()
    expect(() => {
      cpBuilder.addContained(thresholdSet)
    }).to.throw(AlreadyExistsError)
  })
  it('should throw an error when adding a task that is in conflict with an existing task', () => {
    const cpBuilder: CarePlanBuilder = new CarePlanBuilder().use(testCarePlan)
    const task = new TaskBuilder()
      .use(cpBuilder.getContained('task1') as Task)
      .setId('task2')
      .build()
    expect(() => {
      cpBuilder.addContained(task)
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error containing id of conflicting contained resource when adding conflicting resource to contained-array', () => {
    const cpBuilder: CarePlanBuilder = new CarePlanBuilder().use(testCarePlan)
    const task = new TaskBuilder()
      .use(cpBuilder.getContained('task1') as Task)
      .setId('task2')
      .build()
    let error: AlreadyExistsError = new AlreadyExistsError()
    try {
      cpBuilder.addContained(task)
    } catch (e) {
      error = e
    }
    expect(error.conflicts).to.eql(['task2'])
  })

  it("should throw an error containing no id's, if resource doesn't contain an id, when adding conflicting resources to contained-array", () => {
    const cpBuilder: CarePlanBuilder = new CarePlanBuilder().use(testCarePlan)
    const task = new TaskBuilder()
      .use(cpBuilder.getContained('task1') as Task)
      .setId()
      .build()
    let error: AlreadyExistsError = new AlreadyExistsError()
    try {
      cpBuilder.addContained(task)
    } catch (e) {
      error = e
    }
    expect(error.conflicts).to.eql([])
  })

  it('should allow for handling conflicting activities', () => {
    const builder: CarePlanBuilder = new CarePlanBuilder().setActivity([
      new CarePlanActivityBuilder()
        .setId('2')
        .setReference(new ReferenceBuilder().setReference('#ref').build())
        .build(),
    ])

    const activity: CarePlanActivity = new CarePlanActivityBuilder()
      .setId('2')
      .setReference(new ReferenceBuilder().setReference('#ref').build())
      .build()
    try {
      builder.addActivity(activity)
    } catch (e) {
      builder.setActivity([
        ...builder
          .getActivity()
          .filter(act => act !== builder.getConflictingActivity(activity)),
      ])
      builder.addActivity(activity)
    }
    expect(
      builder.getActivity().filter(activity => {
        return new CarePlanActivityBuilder().use(activity).getId() === '2'
      }).length
    ).to.eql(1)
    expect(
      builder.getActivity().filter(activity => {
        return new CarePlanActivityBuilder().use(activity).getId() === '1'
      }).length
    ).to.eql(0)
  })

  it('should allow for handling conflicting contained resources', () => {
    const builder: CarePlanBuilder = new CarePlanBuilder().use(testCarePlan)
    const task = new TaskBuilder()
      .use(builder.getContained('task1') as Task)
      .setId('task2')
      .build()
    try {
      builder.addContained(task)
    } catch (e) {
      builder.setContained([
        ...builder
          .getContained()
          .filter(
            contained => contained !== builder.getConflictingContained(task)
          ),
      ])
      builder.addActivity(task)
    }
    expect(
      builder.getActivity().filter(activity => {
        return new CarePlanActivityBuilder().use(activity).getId() === 'task2'
      }).length
    ).to.eql(1)
    expect(
      builder.getActivity().filter(activity => {
        return new CarePlanActivityBuilder().use(activity).getId() === 'task1'
      }).length
    ).to.eql(0)
  })

  afterEach(() => {
    testCarePlan = undefined
  })

  after(() => {
    sinon.reset()
  })
})
