import { expect } from 'chai'
import { IdentifierBuilder } from '../src/identifier-builder'
import { CodeableConceptBuilder, CodingBuilder } from '../src'
import { AlreadyExistsError } from 'exceptions'

describe('IdentifierBuilder', () => {
  it('should be able to build minimal Identifier', () => {
    expect(new IdentifierBuilder().build()).to.eql({})
  })

  it('should be able to accept existing Identifier', () => {
    expect(
      new IdentifierBuilder()
        .use({
          system: 'urn:oid:1.2.208.176.1.2',
          value: '2512489996',
        })
        .build()
    ).to.eql({
      system: 'urn:oid:1.2.208.176.1.2',
      value: '2512489996',
    })
  })

  it('should be able to reset', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      system: 'urn:oid:1.2.208.176.1.2',
      value: '2512489996',
    })
    expect(identifierBuilder.use().build()).to.eql({})
  })

  it('should be able to set use-field', () => {
    expect(new IdentifierBuilder().setUse('official').build()).to.eql({
      use: 'official',
    })
  })

  it('should be able to unset use-field', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      use: 'official',
    })

    expect(identifierBuilder.setUse().build()).to.eql({})
  })

  it('should be able to return use-field', () => {
    expect(new IdentifierBuilder().use({ use: 'official' }).getUse()).to.eql(
      'official'
    )
  })

  it('should be able to set type-field', () => {
    expect(
      new IdentifierBuilder()
        .setType(
          new CodeableConceptBuilder()
            .addCoding(
              new CodingBuilder()
                .setSystem('http://terminology.hl7.org/CodeSystem/v2-0203')
                .setCode('DL')
                .setDisplay("Driver's license number")
                .build()
            )
            .build()
        )
        .build()
    ).to.eql({
      type: {
        coding: [
          {
            system: 'http://terminology.hl7.org/CodeSystem/v2-0203',
            code: 'DL',
            display: "Driver's license number",
          },
        ],
      },
    })
  })

  it('should be able to unset type-field', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      type: {
        coding: [
          {
            system: 'http://terminology.hl7.org/CodeSystem/v2-0203',
            code: 'DL',
            display: "Driver's license number",
          },
        ],
      },
    })

    expect(identifierBuilder.setType().build()).to.eql({})
  })

  it('should be able to return type-field', () => {
    expect(
      new IdentifierBuilder()
        .use({
          type: {
            coding: [
              {
                system: 'http://terminology.hl7.org/CodeSystem/v2-0203',
                code: 'DL',
                display: "Driver's license number",
              },
            ],
          },
        })
        .getType()
    ).to.eql({
      coding: [
        {
          system: 'http://terminology.hl7.org/CodeSystem/v2-0203',
          code: 'DL',
          display: "Driver's license number",
        },
      ],
    })
  })

  it('should be able to set system-field', () => {
    expect(new IdentifierBuilder().setSystem('system').build()).to.eql({
      system: 'system',
    })
  })

  it('should be able to unset system-field', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      system: 'system',
    })

    expect(identifierBuilder.setSystem().build()).to.eql({})
  })

  it('should be able to return system-field', () => {
    expect(
      new IdentifierBuilder().use({ system: 'system' }).getSystem()
    ).to.eql('system')
  })

  it('should be able to set value-field', () => {
    expect(new IdentifierBuilder().setValue('value').build()).to.eql({
      value: 'value',
    })
  })

  it('should be able to unset value-field', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      value: 'value',
    })

    expect(identifierBuilder.setValue().build()).to.eql({})
  })

  it('should be able to return value-field', () => {
    expect(new IdentifierBuilder().use({ value: 'value' }).getValue()).to.eql(
      'value'
    )
  })

  it('should be able to set period-field', () => {
    expect(
      new IdentifierBuilder()
        .setPeriod({
          start: '1951-06-04',
          end: '1953-08-02',
        })
        .build()
    ).to.eql({
      period: {
        start: '1951-06-04',
        end: '1953-08-02',
      },
    })
  })

  it('should be able to unset period-field', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      period: {
        start: '1951-06-04',
        end: '1953-08-02',
      },
    })

    expect(identifierBuilder.setPeriod().build()).to.eql({})
  })

  it('should be able to return period-field', () => {
    expect(
      new IdentifierBuilder()
        .use({
          period: {
            start: '1951-06-04',
            end: '1953-08-02',
          },
        })
        .getPeriod()
    ).to.eql({
      start: '1951-06-04',
      end: '1953-08-02',
    })
  })

  it('should be able to set assigner-field', () => {
    expect(
      new IdentifierBuilder()
        .setAssigner({
          reference: 'http://example.com/Practitioner/7',
          type: 'Practitioner',
        })
        .build()
    ).to.eql({
      assigner: {
        reference: 'http://example.com/Practitioner/7',
        type: 'Practitioner',
      },
    })
  })

  it('should be able to unset assigner-field', () => {
    const identifierBuilder: IdentifierBuilder = new IdentifierBuilder().use({
      assigner: {
        reference: 'http://example.com/Practitioner/7',
        type: 'Practitioner',
      },
    })

    expect(identifierBuilder.setAssigner().build()).to.eql({})
  })

  it('should be able to return assigner-field', () => {
    expect(
      new IdentifierBuilder()
        .use({
          assigner: {
            reference: 'http://example.com/Practitioner/7',
            type: 'Practitioner',
          },
        })
        .getAssigner()
    ).to.eql({
      reference: 'http://example.com/Practitioner/7',
      type: 'Practitioner',
    })
  })

  it('should be able to tell that two identifiers are in conflict if they have the same system and value', () => {
    expect(
      new IdentifierBuilder()
        .setSystem('system')
        .setValue('value')
        .conflictsWith({
          system: 'system',
          value: 'value',
        })
    ).to.eql(true)
  })

  it('should be able to tell that two identifiers are not in conflict if they have different systems', () => {
    expect(
      new IdentifierBuilder()
        .setSystem('system')
        .setValue('value')
        .conflictsWith({
          system: 'otherSystem',
          value: 'value',
        })
    ).to.eql(false)
  })

  it('should be able to tell that two identifiers are not in conflict if one of them are missing either system or value', () => {
    expect(
      new IdentifierBuilder()
        .setSystem('system')
        .setValue('value')
        .conflictsWith({
          value: 'value',
        })
    ).to.eql(false)
  })
})
