import { expect } from 'chai'
import {
  CarePlan,
  PlanDefinition,
  CarePlanActivity,
  Resource,
  Reference,
  PlanDefinitionAction,
  TaskInput,
} from 'fhir'
import utils from '../src/utils'
import { deepcopy } from './utils'
import sinon from 'sinon'
import {
  PlanDefinitionBuilder,
  CarePlanBuilder,
  PlanDefinitionActionBuilder,
  ExtensionBuilder,
  ThresholdSetBuilder,
  ResourceBuilder,
} from '../src'
import { MissingPropertyError, AlreadyExistsError } from '../src/exceptions'
import careplan from './resources/CarePlan.json'
import plandefinition from './resources/PlanDefinition.json'
import plandefinitionNoTSReference from './resources/PlanDefinition-no-thresholdset-reference.json'
import { Constants } from 'global/constants'

describe('Plan Definition To CarePlan Util', () => {
  let testCarePlan: CarePlan | undefined
  let testPlanDefinition: PlanDefinition | undefined

  before(() => {
    // Mock make to avoid testing generated timestamp
    sinon.stub(utils, '_generateRandomFourCharString').callsFake(() => 'abcd')
    sinon
      .stub(utils, '_generateIsoDate')
      .callsFake(() => '0000-00-00T00:00:00.000Z')
  })

  beforeEach(() => {
    testCarePlan = deepcopy(careplan) as CarePlan
    testPlanDefinition = deepcopy(plandefinition) as PlanDefinition
  })

  describe('addPlanDefinitionContentToCarePlan', () => {
    it('should throw an error, if PlanDefinition is missing either url or version', () => {
      expect(() => {
        const pdBuilder: PlanDefinitionBuilder = new PlanDefinitionBuilder()
          .use(testPlanDefinition)
          .setUrl()
        utils.planDefinitionToCarePlan(
          pdBuilder.build(),
          testCarePlan as CarePlan
        )
      }).to.throw(MissingPropertyError)
    })

    it('should throw an error if PlanDefinition canonical is already in careplan instantiatesCanonical array', () => {
      expect(() => {
        const cpBuilder: CarePlanBuilder = new CarePlanBuilder()
          .use(testCarePlan)
          .addInstantiatesCanonical(
            'canonical://uuid/PlanDefinition/5f60c9f9-5a20-4045-863f-0b0b310c24de|1.0.0'
          )
        utils.planDefinitionToCarePlan(
          testPlanDefinition as PlanDefinition,
          cpBuilder.build()
        )
      }).to.throw(AlreadyExistsError)
    })

    it('should throw an error, if an action contains no thresholdSet extension', () => {
      expect(() => {
        const pdNoThresholSetReference: PlanDefinition = deepcopy(
          plandefinitionNoTSReference
        ) as PlanDefinition
        utils.planDefinitionToCarePlan(
          pdNoThresholSetReference,
          testCarePlan as CarePlan
        )
      }).to.throw(MissingPropertyError)
    })

    it('should throw an error, if ThresholdSet is not in contained array', () => {
      expect(() => {
        const pdBuilder: PlanDefinitionBuilder = new PlanDefinitionBuilder().use(
          testPlanDefinition
        )
        pdBuilder.setContained([
          ...pdBuilder
            .getContained()
            .filter(resource => resource.id !== 'thresholdSet1'),
        ])
        utils.planDefinitionToCarePlan(
          pdBuilder.build(),
          testCarePlan as CarePlan
        )
      }).to.throw(MissingPropertyError)
    })

    it('should throw an error if an action in the PlanDefinition is in conflict with an activity in the CarePlan', () => {
      expect(() => {
        const pdBuilder: PlanDefinitionBuilder = new PlanDefinitionBuilder().use(
          testPlanDefinition
        )
        pdBuilder
          .addAction(
            new PlanDefinitionActionBuilder()
              .addExtension(
                new ExtensionBuilder()
                  .setUrl(Constants.planDefinition.thresholdSet)
                  .setValueReference({
                    reference: '#thresholdSetb',
                  })
                  .build()
              )
              .setDefinitionCanonical(
                'canonical://uuid/Questionnaire/f14997cf-e16e-4e4b-8b3a-1f419386f597|1.0.0'
              )
              .build()
          )
          .addContained(
            new ThresholdSetBuilder()
              .use({
                resourceType: 'Basic',
                id: 'thresholdSetb',
                code: {
                  coding: [
                    {
                      system:
                        'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                      code: 'thresholdset',
                      display: 'ThresholdSet',
                    },
                  ],
                  text: 'ThresholdSet',
                },
              })
              .build()
          )
        utils.planDefinitionToCarePlan(
          pdBuilder.build(),
          testCarePlan as CarePlan
        )
      }).to.throw(AlreadyExistsError)
    })

    it('should throw an error if the action is missing necessary information', () => {
      expect(() => {
        const pdBuilder: PlanDefinitionBuilder = new PlanDefinitionBuilder().use(
          testPlanDefinition
        )
        pdBuilder.setAction([
          new PlanDefinitionActionBuilder()
            .use(pdBuilder.getAction()[0])
            .setDefinitionCanonical()
            .build(),
        ])
        utils.planDefinitionToCarePlan(
          pdBuilder.build(),
          testCarePlan as CarePlan
        )
      }).to.throw(MissingPropertyError)
    })

    it("should add plandefinition url|version to careplan's instantiatesCanonical array ", () => {
      const newCarePlan: CarePlan = utils.planDefinitionToCarePlan(
        testPlanDefinition as PlanDefinition,
        testCarePlan as CarePlan
      )
      expect(newCarePlan.instantiatesCanonical).to.contain(
        `${plandefinition.url}|${plandefinition.version}`
      )
    })

    it("should add plandefinition's contained resources to careplan's contained resources", () => {
      const newCarePlan: CarePlan = utils.planDefinitionToCarePlan(
        testPlanDefinition as PlanDefinition,
        testCarePlan as CarePlan
      )

      expect(newCarePlan.contained).to.contain.deep.members(
        plandefinition.contained
      )
    })

    it('should add an activity to the careplan for each action in the plandefinition', () => {
      const newCarePlan: CarePlan = utils.planDefinitionToCarePlan(
        testPlanDefinition as PlanDefinition,
        testCarePlan as CarePlan
      )
      expect((newCarePlan.activity as CarePlanActivity[]).length).to.eql(2)
    })

    it('should add a task for each action in the PlanDefinition', () => {
      const newCarePlan: CarePlan = utils.planDefinitionToCarePlan(
        testPlanDefinition as PlanDefinition,
        testCarePlan as CarePlan
      )
      const tasks = (newCarePlan.contained as Resource[]).filter(
        resource => resource.resourceType === 'Task'
      )
      expect(tasks.length).to.eql(
        careplan.activity.length + plandefinition.action.length
      )
    })

    it("each task in the CarePlan's contained array should be referenced by an activity", () => {
      const newCarePlan: CarePlan = utils.planDefinitionToCarePlan(
        testPlanDefinition as PlanDefinition,
        testCarePlan as CarePlan
      )
      const tasksInContainedArray = (newCarePlan.contained as Resource[])
        .filter(resource => resource.resourceType === 'Task')
        .map(resource => {
          return resource.id
        })

      const referencesToTasksInActivityArray = (newCarePlan.activity as CarePlanActivity[]).map(
        activity => {
          const ref = (activity.reference as Reference).reference as string
          return ref.substr(1, ref.length - 1)
        }
      )

      expect(JSON.stringify(referencesToTasksInActivityArray)).to.eql(
        JSON.stringify(tasksInContainedArray)
      )
    })
  })

  describe('actionToActivity', () => {
    it('should throw an error when action is missing definitionCanonical property', () => {
      expect(() => {
        utils.actionToActivity(
          new PlanDefinitionActionBuilder()
            .addExtension(
              new ExtensionBuilder()
                .use({
                  url: Constants.planDefinition.thresholdSet,
                  valueReference: {
                    reference: '#thresholdSet10',
                  },
                })
                .build()
            )
            .build(),
          new ThresholdSetBuilder().build(),
          new CarePlanBuilder().build()
        )
      }).to.throw(MissingPropertyError)
    })

    it('should throw an error if the definitionCanonical in the action, is referencing to a resource, already referenced by a task in the CarePlan', () => {
      const thresholdSet = new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          id: 'thresholdSet10',
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              valueCanonical:
                'canonical://uuid/Questionnaire/08c9a37d-fd0b-456d-bd8c-81bb6ea3d7e8|1.0.0',
            },
          ],
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .build()

      const action = new PlanDefinitionActionBuilder()
        // conflicting canonical reference
        .setDefinitionCanonical(
          'canonical://uuid/Questionnaire/f14997cf-e16e-4e4b-8b3a-1f419386f597|1.0.0'
        )
        .addExtension(
          new ExtensionBuilder()
            .use({
              url: Constants.planDefinition.thresholdSet,
              valueReference: {
                reference: '#thresholdSet10',
              },
            })
            .build()
        )
        .build()

      expect(() => {
        utils.actionToActivity(action, thresholdSet, testCarePlan as CarePlan)
      }).to.throw(AlreadyExistsError)
    })

    it('should throw an error if the thresholdsSet is in conflict with a thresholdSet in the CarePlan', () => {
      const thresholdSet = new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          id: 'thresholdSet10',
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              // Conflicting canonical reference
              valueCanonical:
                'canonical://uuid/Questionnaire/f14997cf-e16e-4e4b-8b3a-1f419386f597|1.0.0',
            },
          ],
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .build()

      const action = new PlanDefinitionActionBuilder()
        .setDefinitionCanonical(
          'canonical://uuid/Questionnaire/08c9a37d-fd0b-456d-bd8c-81bb6ea3d7e8|1.0.0'
        )
        .addExtension(
          new ExtensionBuilder()
            .use({
              url: Constants.planDefinition.thresholdSet,
              valueReference: {
                reference: '#thresholdSet10',
              },
            })
            .build()
        )
        .build()

      expect(() => {
        utils.actionToActivity(action, thresholdSet, testCarePlan as CarePlan)
      }).to.throw(AlreadyExistsError)
    })

    it('should add new activity to CarePlan, if no errors are thrown', () => {
      const thresholdSet = new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          id: 'thresholdSet10',
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              // Conflicting canonical reference
              valueCanonical:
                'canonical://uuid/Questionnaire/08c9a37d-fd0b-456d-bd8c-81bb6ea3d7e8|1.0.0',
            },
          ],
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .build()

      const action = new PlanDefinitionActionBuilder()
        .setDefinitionCanonical(
          'canonical://uuid/Questionnaire/08c9a37d-fd0b-456d-bd8c-81bb6ea3d7e8|1.0.0'
        )
        .addExtension(
          new ExtensionBuilder()
            .use({
              url: Constants.planDefinition.thresholdSet,
              valueReference: {
                reference: '#thresholdSet10',
              },
            })
            .build()
        )
        .build()
      const carePlan = utils.actionToActivity(
        action,
        thresholdSet,
        testCarePlan as CarePlan
      )
      expect((carePlan.activity as CarePlanActivity[]).length).to.eql(2)
    })
  })

  describe('createTask', () => {
    it('should create a task containing the provided canonical-reference', () => {
      const task = utils._createTask(
        'canonical://uuid/Questionnaire/08c9a37d-fd0b-456d-bd8c-81bb6ea3d7e8|1.0.0'
      )
      expect(task).to.have.property('resourceType', 'Task')
      expect(task).to.have.property('status', 'requested')
      expect(task).to.have.property('intent', 'order')
      expect(task.input).to.eql([
        {
          type: {
            coding: [
              {
                system: 'http://hl7.org/fhir/resource-types',
                code: 'Questionnaire',
                display: 'Questionnaire',
              },
            ],
            text: 'Questionnaire',
          },
          valueCanonical:
            'canonical://uuid/Questionnaire/08c9a37d-fd0b-456d-bd8c-81bb6ea3d7e8|1.0.0',
        },
      ])
    })
  })

  describe('handleIdConflicts', () => {
    it('should return the same resource if there are no id conflicts', () => {
      const cpBuilder = new CarePlanBuilder().use(testCarePlan)
      const resourceBuilder = new ResourceBuilder().use(
        cpBuilder.getContained()[0]
      )
      const resource = utils._handleIdConflicts(
        resourceBuilder.setId('non-conflicting-id').build(),
        cpBuilder.build()
      )
      expect(resource).to.eql(resourceBuilder.build())
    })

    it('should generate new Id for resource if there is a conflict', () => {
      const cpBuilder = new CarePlanBuilder().use(testCarePlan)
      const resourceBuilder = new ResourceBuilder().use(
        cpBuilder.getContained()[0]
      )
      const resource = utils._handleIdConflicts(
        resourceBuilder.build(),
        cpBuilder.build()
      )
      expect(resource.id).to.not.eql(resourceBuilder.getId())
    })
  })

  describe('getThresholdSetReference', () => {
    it('should return thresholdSet reference if one exists', () => {
      const pdActionBuilder = new PlanDefinitionActionBuilder().use(
        new PlanDefinitionBuilder().use(testPlanDefinition).getAction()[0]
      )

      expect(utils._getThresholdSetReference(pdActionBuilder)).to.eql(
        '#thresholdSet1'
      )
    })

    it('should return undefined if no thresholdSet reference exists', () => {
      const pdActionBuilder = new PlanDefinitionActionBuilder()
        .use(new PlanDefinitionBuilder().use(testPlanDefinition).getAction()[0])
        .setExtension([])

      expect(utils._getThresholdSetReference(pdActionBuilder)).to.be.an(
        'undefined'
      )
    })
  })

  describe('createActivity', () => {
    it('should create an activity if all necessary information is provided', () => {
      expect(utils._createActivity('taskId', 'thresholdSetId')).to.eql({
        reference: {
          reference: '#taskId',
          type: 'Task',
        },
        extension: [
          {
            url: Constants.carePlan.thresholdSet,
            valueReference: {
              reference: '#thresholdSetId',
            },
          },
        ],
      })
    })

    it('should throw an error if no taskId provided', () => {
      expect(() => {
        utils._createActivity(undefined, 'thresholdSetId')
      }).to.throw(MissingPropertyError)
    })
    it('should throw an error if no thresholdSetId provided', () => {
      expect(() => {
        utils._createActivity('taskId', undefined)
      }).to.throw(MissingPropertyError)
    })
  })

  afterEach(() => {
    testCarePlan = undefined
    testPlanDefinition = undefined
  })

  after(() => {
    sinon.reset()
  })
})
