import { expect } from 'chai'
import { ResourceBuilder } from '../src'
import { Resource } from 'fhir'

describe('ResourceBuilder', () => {
  it('should be able to build minimal Resource', () => {
    expect(new ResourceBuilder().build()).to.eql({ resourceType: 'Resource' })
  })

  it('should be able to accept existing Resource', () => {
    expect(
      new ResourceBuilder()
        .use({
          resourceType: 'Resource',
          id: 'resource-7',
        })
        .build()
    ).to.eql({
      resourceType: 'Resource',
      id: 'resource-7',
    })
  })

  it('should be able to reset', () => {
    const planDefinitionActionBuilder: ResourceBuilder = new ResourceBuilder().use(
      {
        resourceType: 'Resource',
        id: 'resource-7',
      }
    )
    expect(planDefinitionActionBuilder.use().build()).to.eql({
      resourceType: 'Resource',
    })
  })

  it('should be able to set resourceType-field', () => {
    expect(new ResourceBuilder().setResourceType('Resource').build()).to.eql({
      resourceType: 'Resource',
    })
  })

  it('should be able to unset resourceType-field', () => {
    const resourceBuilder: ResourceBuilder = new ResourceBuilder().use({
      resourceType: 'Resource',
    })

    expect(resourceBuilder.setResourceType().build()).to.eql({})
  })

  it('should be able to return resourceType-field', () => {
    expect(
      new ResourceBuilder().use({ resourceType: 'Resource' }).getResourceType()
    ).to.eql('Resource')
  })

  it('should be able to set id-field', () => {
    expect(new ResourceBuilder().setId('id').build()).to.eql({
      resourceType: 'Resource',
      id: 'id',
    })
  })

  it('should be able to unset id-field', () => {
    const resourceBuilder: ResourceBuilder = new ResourceBuilder().use({
      resourceType: 'Resource',
      id: 'id',
    })

    expect(resourceBuilder.setId().build()).to.eql({
      resourceType: 'Resource',
    })
  })

  it('should be able to return id-field', () => {
    expect(new ResourceBuilder().use({ id: 'id' }).getId()).to.eql('id')
  })

  it('should be able to set meta-field', () => {
    expect(
      new ResourceBuilder()
        .setMeta({
          lastUpdated: '2019-06-24T14:00:00.000Z',
          profile: ['http://hl7.org/fhir/uv/phd/StructureDefinition/PhgDevice'],
        })
        .build()
    ).to.eql({
      resourceType: 'Resource',
      meta: {
        lastUpdated: '2019-06-24T14:00:00.000Z',
        profile: ['http://hl7.org/fhir/uv/phd/StructureDefinition/PhgDevice'],
      },
    })
  })

  it('should be able to unset meta-field', () => {
    const resourceBuilder: ResourceBuilder = new ResourceBuilder().use({
      resourceType: 'Resource',
      meta: {
        lastUpdated: '2019-06-24T14:00:00.000Z',
        profile: ['http://hl7.org/fhir/uv/phd/StructureDefinition/PhgDevice'],
      },
    })

    expect(resourceBuilder.setMeta().build()).to.eql({
      resourceType: 'Resource',
    })
  })

  it('should be able to return meta-field', () => {
    expect(
      new ResourceBuilder()
        .use({
          meta: {
            lastUpdated: '2019-06-24T14:00:00.000Z',
            profile: [
              'http://hl7.org/fhir/uv/phd/StructureDefinition/PhgDevice',
            ],
          },
        })
        .getMeta()
    ).to.eql({
      lastUpdated: '2019-06-24T14:00:00.000Z',
      profile: ['http://hl7.org/fhir/uv/phd/StructureDefinition/PhgDevice'],
    })
  })

  it('should be able to set implicitRules-field', () => {
    expect(
      new ResourceBuilder().setImplicitRules('http://example.com/rules').build()
    ).to.eql({
      resourceType: 'Resource',
      implicitRules: 'http://example.com/rules',
    })
  })

  it('should be able to unset implicitRules-field', () => {
    const resourceBuilder: ResourceBuilder = new ResourceBuilder().use({
      resourceType: 'Resource',
      implicitRules: 'http://example.com/rules',
    })

    expect(resourceBuilder.setImplicitRules().build()).to.eql({
      resourceType: 'Resource',
    })
  })

  it('should be able to return implicitRules-field', () => {
    expect(
      new ResourceBuilder()
        .use({ implicitRules: 'http://example.com/rules' })
        .getImplicitRules()
    ).to.eql('http://example.com/rules')
  })

  it('should be able to set language-field', () => {
    expect(new ResourceBuilder().setLanguage('da').build()).to.eql({
      resourceType: 'Resource',
      language: 'da',
    })
  })

  it('should be able to unset language-field', () => {
    const resourceBuilder: ResourceBuilder = new ResourceBuilder().use({
      resourceType: 'Resource',
      language: 'da',
    })

    expect(resourceBuilder.setLanguage().build()).to.eql({
      resourceType: 'Resource',
    })
  })

  it('should be able to return language-field', () => {
    expect(new ResourceBuilder().use({ language: 'da' }).getLanguage()).to.eql(
      'da'
    )
  })

  it('should be able to tell that two resources are in conflict, if they have the same id', () => {
    expect(
      new ResourceBuilder()
        .setId('id')
        .conflictsWith(new ResourceBuilder().setId('id').build())
    ).to.eql(true)
  })

  it('should be able to tell that two resources are not in conflict, if they have different ids', () => {
    expect(
      new ResourceBuilder()
        .setId('id')
        .conflictsWith(new ResourceBuilder().setId('id2').build())
    ).to.eql(false)
  })
})
