import { expect } from 'chai'
import {
  TaskBuilder,
  ReferenceBuilder,
  TaskInputBuilder,
  CodeableConceptBuilder,
  CodingBuilder,
} from '../src'
import { Task, Identifier, Reference, TaskInput } from 'fhir'
import { AlreadyExistsError, MissingPropertyError } from '../src/exceptions'

describe('TaskBuilder', () => {
  it('should be able to build minimal Task', () => {
    expect(new TaskBuilder().build()).to.eql({
      resourceType: 'Task',
    })
  })

  it('should be able to accept existing Task', () => {
    expect(
      new TaskBuilder()
        .use({
          resourceType: 'Task',
          status: 'requested',
          intent: 'plan',
        })
        .build()
    ).to.eql({
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
    })
  })

  it('should be able to reset', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().use({
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
    })
    expect(taskBuilder.use().build()).to.eql({
      resourceType: 'Task',
    })
  })

  it('should be able to set identifiers for a task', () => {
    const task: Task = new TaskBuilder()
      .setIdentifier([
        {
          system: 'system',
          value: 'value',
        },
      ])
      .build()
    expect(task).to.eql({
      resourceType: 'Task',
      identifier: [
        {
          system: 'system',
          value: 'value',
        },
      ],
    })
  })

  it('should be able to unset identifiers for a task', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().use({
      resourceType: 'Task',
      identifier: [
        {
          system: 'system',
          value: 'value',
        },
      ],
    } as Task)

    expect(taskBuilder.setIdentifier().build()).to.eql({
      resourceType: 'Task',
    })
  })

  it('should be able to add identifier to a task', () => {
    const task: Task = new TaskBuilder()
      .addIdentifier({
        system: 'system',
        value: 'value',
      })
      .build()
    expect(task.identifier).to.eql([
      {
        system: 'system',
        value: 'value',
      },
    ])
  })

  it('should throw an error when adding an identifier in a system in which there is already an identifier', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().addIdentifier({
      system: 'existing',
      value: 'existing',
    })
    expect(() => {
      taskBuilder.addIdentifier({
        system: 'existing',
        value: 'existing',
      })
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error when adding an identifier system or value', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder()
    expect(() => {
      taskBuilder.addIdentifier({
        value: 'existing',
      })
    }).to.throw(MissingPropertyError)
  })

  it('should return empty array when getting identifiers and none defined', () => {
    const identifiers: Identifier[] = new TaskBuilder().getIdentifier()
    expect(identifiers).to.eql([])
  })

  it('should be able to return array of identifiers', () => {
    const identifiers: Identifier[] = new TaskBuilder()
      .setIdentifier([
        {
          system: 'system',
          value: 'value',
        },
      ])
      .getIdentifier()
    expect(identifiers).to.eql([
      {
        system: 'system',
        value: 'value',
      },
    ])
  })

  it("should return undefined when getting identifier in system that isn't there", () => {
    const identifier: Identifier | undefined = new TaskBuilder().getIdentifier(
      'none existent system'
    )
    expect(identifier).to.eql(undefined)
  })

  it('should be able return an identifier based on system', () => {
    const identifier: Identifier | undefined = new TaskBuilder()
      .setIdentifier([
        {
          system: 'system',
          value: 'value',
        },
      ])
      .getIdentifier('system')
    expect(identifier).to.eql({ system: 'system', value: 'value' })
  })

  it('should be able to set basedOn-array', () => {
    const reference: Reference = new ReferenceBuilder()
      .setReference('http://example.com/CarePlan/7')
      .build()

    const task: Task = new TaskBuilder().setBasedOn([reference]).build()

    expect(task).to.eql({
      resourceType: 'Task',
      basedOn: [
        {
          reference: 'http://example.com/CarePlan/7',
        },
      ],
    })
  })

  it('should be able to unset basedOn-array', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().use({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
      basedOn: [
        {
          reference: 'http://example.com/CarePlan/7',
        },
      ],
    })

    expect(taskBuilder.setBasedOn().build()).to.eql({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
    })
  })

  it('should throw an error when setting basedOn-array containing an item without necessary information', () => {
    expect(() => {
      new TaskBuilder().setBasedOn([
        {
          identifier: {
            system: 'system',
          },
        },
      ])
    }).to.throw(MissingPropertyError)
  })

  it('should be able to return basedOn-array', () => {
    const task: Task = {
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
      basedOn: [{ reference: 'http://example.com/CarePlan/7' }],
    }

    expect(new TaskBuilder().use(task).getBasedOn()).to.eql([
      { reference: 'http://example.com/CarePlan/7' },
    ])
  })

  it('should return empty array when basedOn is empty', () => {
    expect(new TaskBuilder().getBasedOn()).to.eql([])
  })

  it('should be able to add reference to basedOn-array', () => {
    expect(
      new TaskBuilder()
        .use({
          resourceType: 'Task',
          status: 'requested',
          intent: 'plan',
          basedOn: [{ reference: 'http://example.com/CarePlan/8' }],
        })
        .addBasedOn({ reference: 'http://example.com/CarePlan/7' })
        .getBasedOn()
    ).to.eql([
      { reference: 'http://example.com/CarePlan/8' },
      { reference: 'http://example.com/CarePlan/7' },
    ])
  })

  it('should throw an error when adding an item with a reference that already exists in the basedOn-array', () => {
    const task: Task = {
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
      basedOn: [{ reference: 'http://example.com/CarePlan/7' }],
    }
    expect(() => {
      new TaskBuilder()
        .use(task)
        .addBasedOn({ reference: 'http://example.com/CarePlan/7' })
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error when adding an item with an identifier that already exists in the basedOn-array', () => {
    const task: Task = {
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
      basedOn: [
        {
          identifier: {
            system: 'system',
            value: 'value',
          },
        },
      ],
    }
    expect(() => {
      new TaskBuilder().use(task).addBasedOn({
        identifier: {
          system: 'system',
          value: 'value',
        },
      })
    }).to.throw(AlreadyExistsError)
  })

  it('should throw an error when adding an item without necessary information to the basedOn-array', () => {
    expect(() => {
      new TaskBuilder().addBasedOn({
        identifier: {
          system: 'system',
        },
      })
    }).to.throw(MissingPropertyError)
  })

  it('should be able to set status-field', () => {
    expect(new TaskBuilder().setStatus('in-progress').build()).to.eql({
      resourceType: 'Task',
      status: 'in-progress',
    })
  })

  it('should be able to return status-field', () => {
    expect(
      new TaskBuilder().use({ status: 'draft' } as Task).getStatus()
    ).to.eql('draft')
  })

  it('should be able to set intent-field', () => {
    expect(new TaskBuilder().setIntent('option').build()).to.eql({
      resourceType: 'Task',
      intent: 'option',
    })
  })

  it('should be able to return intent-field', () => {
    expect(
      new TaskBuilder().use({ intent: 'option' } as Task).getIntent()
    ).to.eql('option')
  })

  it('should be able to set authoredOn-field', () => {
    expect(
      new TaskBuilder().setAuthoredOn('2013-06-08T10:57:34+01:00').build()
    ).to.eql({
      resourceType: 'Task',
      authoredOn: '2013-06-08T10:57:34+01:00',
    })
  })

  it('should be able to set authoredOn-field', () => {
    expect(
      new TaskBuilder().setAuthoredOn('2013-06-08T10:57:34+01:00').build()
    ).to.eql({
      resourceType: 'Task',
      authoredOn: '2013-06-08T10:57:34+01:00',
    })
  })

  it('should be able to unset authoredOn-field', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().use({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
      authoredOn: '2013-06-08T10:57:34+01:00',
    })

    expect(taskBuilder.setAuthoredOn().build()).to.eql({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
    })
  })

  it('should be able to return authoredOn-field', () => {
    expect(
      new TaskBuilder()
        .use({ authoredOn: '2013-06-08T10:57:34+01:00' } as Task)
        .getAuthoredOn()
    ).to.eql('2013-06-08T10:57:34+01:00')
  })

  it('should be able to return authoredOn-field', () => {
    expect(
      new TaskBuilder()
        .use({ authoredOn: '2013-06-08T10:57:34+01:00' } as Task)
        .getAuthoredOn()
    ).to.eql('2013-06-08T10:57:34+01:00')
  })

  it('should be able to set lastModified-field', () => {
    expect(
      new TaskBuilder().setLastModified('2013-06-08T10:57:34+01:00').build()
    ).to.eql({
      resourceType: 'Task',
      lastModified: '2013-06-08T10:57:34+01:00',
    })
  })

  it('should be able to unset lastModified-field', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().use({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
      lastModified: '2013-06-08T10:57:34+01:00',
    })

    expect(taskBuilder.setLastModified().build()).to.eql({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
    })
  })

  it('should be able to return lastModified-field', () => {
    expect(
      new TaskBuilder()
        .use({ lastModified: '2013-06-08T10:57:34+01:00' } as Task)
        .getLastModified()
    ).to.eql('2013-06-08T10:57:34+01:00')
  })

  it('should be able to set input-array', () => {
    expect(
      new TaskBuilder()
        .setInput([
          new TaskInputBuilder()
            .setId('id')
            .setType(
              new CodeableConceptBuilder()
                .setCoding([
                  new CodingBuilder()
                    .setSystem('system')
                    .setCode('code')
                    .build(),
                ])
                .build()
            )
            .build(),
        ])
        .build()
    ).to.eql({
      resourceType: 'Task',
      input: [
        {
          id: 'id',
          type: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
          },
        },
      ],
    })
  })

  it('should be able to unset input-array', () => {
    const taskBuilder: TaskBuilder = new TaskBuilder().use({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
      input: [
        {
          id: 'id',
          type: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
          },
        },
      ],
    })

    expect(taskBuilder.setInput().build()).to.eql({
      resourceType: 'Task',
      status: 'requested',
      intent: 'order',
    })
  })

  it('should be able to return basedOn-array', () => {
    const task: Task = {
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
      input: [
        {
          id: 'id',
          type: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
          },
        },
      ],
    }

    expect(new TaskBuilder().use(task).getInput()).to.eql([
      {
        id: 'id',
        type: {
          coding: [
            {
              system: 'system',
              code: 'code',
            },
          ],
        },
      },
    ])
  })

  it('should return empty array when basedOn is empty', () => {
    expect(new TaskBuilder().getInput()).to.eql([])
  })

  it('should be able to add input to input-array', () => {
    expect(
      new TaskBuilder()
        .use({
          resourceType: 'Task',
          status: 'requested',
          intent: 'plan',
        })
        .addInput({
          id: 'id',
          type: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
          },
        })
        .getInput()
    ).to.eql([
      {
        id: 'id',
        type: {
          coding: [
            {
              system: 'system',
              code: 'code',
            },
          ],
        },
      },
    ])
  })

  it('should throw an error when adding an input that conflicts with an input that already exists in the input-array', () => {
    const task: Task = {
      resourceType: 'Task',
      status: 'requested',
      intent: 'plan',
      input: [
        {
          id: 'id',
          type: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
          },
          valueCanonical:
            'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
        },
      ],
    }
    expect(() => {
      new TaskBuilder().use(task).addInput({
        id: 'id',
        type: {
          coding: [
            {
              system: 'system',
              code: 'code',
            },
          ],
        },
        valueCanonical:
          'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
      })
    }).to.throw(AlreadyExistsError)
  })

  it('should be able to tell that two tasks are in conflict, if they reference the same resource', () => {
    expect(
      new TaskBuilder()
        .addInput(
          new TaskInputBuilder()
            .setValueCanonical(
              'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0'
            )
            .build()
        )
        .conflictsWith(
          new TaskBuilder()
            .addInput(
              new TaskInputBuilder()
                .setValueCanonical(
                  'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0'
                )
                .build()
            )
            .build()
        )
    ).to.eql(true)
  })

  it('should be able to tell that two tasks are not in conflict, if they reference the different resources', () => {
    expect(
      new TaskBuilder()
        .addInput(
          new TaskInputBuilder()
            .setValueCanonical(
              'canonical://uuid/Questionnaire/39268cb0-efb0-4e17-9b5b-9dfc7799495c|1.0.0'
            )
            .build()
        )
        .conflictsWith(
          new TaskBuilder()
            .addInput(
              new TaskInputBuilder()
                .setValueCanonical(
                  'canonical://uuid/Questionnaire/fd137ea4-da35-44ec-9fb2-fae3dc864506|1.0.0'
                )
                .build()
            )
            .build()
        )
    ).to.eql(false)
  })
})
