import { expect } from 'chai'
import { ThresholdEffectiveWhenBuilder } from 'threshold-effective-when-builder'
import { Constants } from 'global/constants'
import { MalformedError } from 'exceptions/malformed-error'

describe('ThresholdEffectiveWhenBuilder', () => {
  it('should be able to build minimal effectiveWhen', () => {
    expect(new ThresholdEffectiveWhenBuilder().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to accept existing threshold', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.operator,
              valueCode: '<',
            },
          ],
        })
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.operator,
          valueCode: '<',
        },
      ],
    })
  })

  it('should be able to reset', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.operator,
            valueCode: '<',
          },
        ],
      }
    )
    expect(thresholdBuilder.use().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to set component-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setComponent({
          system:
            'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
          code: 'systolic',
        })
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.component,
          valueCoding: {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
            code: 'systolic',
          },
        },
      ],
    })
  })

  it('should be able to unset component-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.component,
            valueCoding: {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
              code: 'systolic',
            },
          },
        ],
      }
    )

    expect(thresholdBuilder.setComponent().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return component-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.component,
              valueCoding: {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
                code: 'systolic',
              },
            },
          ],
        })
        .getComponent()
    ).to.eql({
      system:
        'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
      code: 'systolic',
    })
  })

  it('should throw error if component-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.component,
              valueCoding: {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
                code: 'systolic',
              },
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.component,
          valueCoding: {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100433987/ThresholdSet+component+coding+system',
            code: 'systolic',
          },
        })
        .getComponent()
    }).to.throw(MalformedError)
  })

  it('should return undefined if component-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getComponent()
    ).to.be.an('undefined')
  })

  it('should be able to set valueBoolean-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder().setValueBoolean(true).build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueBoolean,
          valueBoolean: true,
        },
      ],
    })
  })

  it('should be able to unset valueBoolean-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueBoolean,
            valueBoolean: true,
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueBoolean().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueBoolean-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueBoolean,
              valueBoolean: true,
            },
          ],
        })
        .getValueBoolean()
    ).to.eql(true)
  })

  it('should throw error if valueBoolean-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueBoolean,
              valueBoolean: true,
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueBoolean,
          valueBoolean: true,
        })
        .getValueBoolean()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueBoolean-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueBoolean()
    ).to.be.an('undefined')
  })

  it('should be able to set valueDecimal-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder().setValueDecimal(2.24).build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueDecimal,
          valueDecimal: 2.24,
        },
      ],
    })
  })

  it('should be able to unset valueDecimal-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueDecimal,
            valueDecimal: 2.24,
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueDecimal().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueDecimal-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueDecimal,
              valueDecimal: 2.24,
            },
          ],
        })
        .getValueDecimal()
    ).to.eql(2.24)
  })

  it('should throw error if valueDecimal-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueDecimal,
              valueDecimal: 2.24,
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueDecimal,
          valueDecimal: 2.24,
        })
        .getValueDecimal()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueDecimal-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueDecimal()
    ).to.be.an('undefined')
  })

  it('should be able to set valueInteger-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder().setValueInteger(7).build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueInteger,
          valueInteger: 7,
        },
      ],
    })
  })

  it('should be able to unset valueInteger-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueInteger,
            valueInteger: 7,
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueInteger().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueInteger-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueInteger,
              valueInteger: 7,
            },
          ],
        })
        .getValueInteger()
    ).to.eql(7)
  })

  it('should throw error if valueInteger-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueInteger,
              valueInteger: 7,
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueInteger,
          valueInteger: 7,
        })
        .getValueInteger()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueInteger-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueInteger()
    ).to.be.an('undefined')
  })

  it('should be able to set valueString-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder().setValueString('String').build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueString,
          valueString: 'String',
        },
      ],
    })
  })

  it('should be able to unset valueString-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueString,
            valueString: 'String',
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueString().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueString-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueString,
              valueString: 'String',
            },
          ],
        })
        .getValueString()
    ).to.eql('String')
  })

  it('should throw error if valueString-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueString,
              valueString: 'String',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueString,
          valueString: 'String',
        })
        .getValueString()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueString-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueString()
    ).to.be.an('undefined')
  })

  it('should be able to set valueDate-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder().setValueDate('1951-06-04').build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueDate,
          valueDate: '1951-06-04',
        },
      ],
    })
  })

  it('should be able to unset valueDate-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueDate,
            valueDate: '1951-06-04',
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueDate().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueDate-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueDate,
              valueDate: '1951-06-04',
            },
          ],
        })
        .getValueDate()
    ).to.eql('1951-06-04')
  })

  it('should throw error if valueDate-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueDate,
              valueDate: '1951-06-04',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueDate,
          valueDate: '1951-06-04',
        })
        .getValueDate()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueDate-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueDate()
    ).to.be.an('undefined')
  })

  it('should be able to set valueDateTime-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setValueDateTime('2013-06-08T10:57:34+01:00')
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueDateTime,
          valueDateTime: '2013-06-08T10:57:34+01:00',
        },
      ],
    })
  })

  it('should be able to unset valueDateTime-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueDateTime,
            valueDateTime: '2013-06-08T10:57:34+01:00',
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueDateTime().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueDateTime-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueDateTime,
              valueDateTime: '2013-06-08T10:57:34+01:00',
            },
          ],
        })
        .getValueDateTime()
    ).to.eql('2013-06-08T10:57:34+01:00')
  })

  it('should throw error if valueDateTime-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueDateTime,
              valueDateTime: '2013-06-08T10:57:34+01:00',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueDateTime,
          valueDateTime: '2013-06-08T10:57:34+01:00',
        })
        .getValueDateTime()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueDateTime-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueDateTime()
    ).to.be.an('undefined')
  })

  it('should be able to set valueUri-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setValueUri('urn:iso:std:iso:11073:10101')
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueUri,
          valueUri: 'urn:iso:std:iso:11073:10101',
        },
      ],
    })
  })

  it('should be able to unset valueUri-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueUri,
            valueUri: 'urn:iso:std:iso:11073:10101',
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueUri().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueUri-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueUri,
              valueUri: 'urn:iso:std:iso:11073:10101',
            },
          ],
        })
        .getValueUri()
    ).to.eql('urn:iso:std:iso:11073:10101')
  })

  it('should throw error if valueUri-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueUri,
              valueUri: 'urn:iso:std:iso:11073:10101',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueUri,
          valueUri: 'urn:iso:std:iso:11073:10101',
        })
        .getValueUri()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueUri-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueUri()
    ).to.be.an('undefined')
  })

  it('should be able to set valueAttachment-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setValueAttachment({
          title: 'Attachment',
        })
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueAttachment,
          valueAttachment: {
            title: 'Attachment',
          },
        },
      ],
    })
  })

  it('should be able to unset valueAttachment-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueAttachment,
            valueAttachment: {
              title: 'Attachment',
            },
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueAttachment().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueAttachment-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueAttachment,
              valueAttachment: {
                title: 'Attachment',
              },
            },
          ],
        })
        .getValueAttachment()
    ).to.eql({
      title: 'Attachment',
    })
  })

  it('should throw error if valueAttachment-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueAttachment,
              valueAttachment: {
                title: 'Attachment',
              },
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueAttachment,
          valueAttachment: {
            title: 'Attachment',
          },
        })
        .getValueAttachment()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueAttachment-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueAttachment()
    ).to.be.an('undefined')
  })

  it('should be able to set valueCodeableConcept-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setValueCodeableConcept({
          coding: [
            {
              system: 'system',
              code: 'code',
            },
          ],
          text: 'CodeableConcept',
        })
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueCodeableConcept,
          valueCodeableConcept: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
            text: 'CodeableConcept',
          },
        },
      ],
    })
  })

  it('should be able to unset valueCodeableConcept-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueCodeableConcept,
            valueCodeableConcept: {
              coding: [
                {
                  system: 'system',
                  code: 'code',
                },
              ],
              text: 'CodeableConcept',
            },
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueCodeableConcept().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueCodeableConcept-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueCodeableConcept,
              valueCodeableConcept: {
                coding: [
                  {
                    system: 'system',
                    code: 'code',
                  },
                ],
                text: 'CodeableConcept',
              },
            },
          ],
        })
        .getValueCodeableConcept()
    ).to.eql({
      coding: [
        {
          system: 'system',
          code: 'code',
        },
      ],
      text: 'CodeableConcept',
    })
  })

  it('should throw error if valueCodeableConcept-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueCodeableConcept,
              valueCodeableConcept: {
                coding: [
                  {
                    system: 'system',
                    code: 'code',
                  },
                ],
                text: 'CodeableConcept',
              },
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueCodeableConcept,
          valueCodeableConcept: {
            coding: [
              {
                system: 'system',
                code: 'code',
              },
            ],
            text: 'CodeableConcept',
          },
        })
        .getValueCodeableConcept()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueCodeableConcept-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueCodeableConcept()
    ).to.be.an('undefined')
  })

  it('should be able to set valueQuantity-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setValueQuantity({
          value: 250.3,
          system: 'http://unitsofmeasure.org ',
          code: 'kg',
        })
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueQuantity,
          valueQuantity: {
            value: 250.3,
            system: 'http://unitsofmeasure.org ',
            code: 'kg',
          },
        },
      ],
    })
  })

  it('should be able to unset valueQuantity-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueQuantity,
            valueQuantity: {
              value: 250.3,
              system: 'http://unitsofmeasure.org ',
              code: 'kg',
            },
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueQuantity().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueQuantity-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueQuantity,
              valueQuantity: {
                value: 250.3,
                system: 'http://unitsofmeasure.org ',
                code: 'kg',
              },
            },
          ],
        })
        .getValueQuantity()
    ).to.eql({
      value: 250.3,
      system: 'http://unitsofmeasure.org ',
      code: 'kg',
    })
  })

  it('should throw error if valueQuantity-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueQuantity,
              valueQuantity: {
                value: 250.3,
                system: 'http://unitsofmeasure.org ',
                code: 'kg',
              },
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueQuantity,
          valueQuantity: {
            value: 250.3,
            system: 'http://unitsofmeasure.org ',
            code: 'kg',
          },
        })
        .getValueQuantity()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueQuantity-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueQuantity()
    ).to.be.an('undefined')
  })

  it('should be able to set valueReference-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .setValueReference({
          reference: '#task1',
          type: 'Task',
        })
        .build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.valueReference,
          valueReference: {
            reference: '#task1',
            type: 'Task',
          },
        },
      ],
    })
  })

  it('should be able to unset valueReference-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.valueReference,
            valueReference: {
              reference: '#task1',
              type: 'Task',
            },
          },
        ],
      }
    )

    expect(thresholdBuilder.setValueReference().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return valueReference-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueReference,
              valueReference: {
                reference: '#task1',
                type: 'Task',
              },
            },
          ],
        })
        .getValueReference()
    ).to.eql({
      reference: '#task1',
      type: 'Task',
    })
  })

  it('should throw error if valueReference-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.valueReference,
              valueReference: {
                reference: '#task1',
                type: 'Task',
              },
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.valueReference,
          valueReference: {
            reference: '#task1',
            type: 'Task',
          },
        })
        .getValueReference()
    }).to.throw(MalformedError)
  })

  it('should return undefined if valueReference-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getValueReference()
    ).to.be.an('undefined')
  })

  it('should be able to set operator-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder().setOperator('<=').build()
    ).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [
        {
          url: Constants.thresholdSet.operator,
          valueCode: '<=',
        },
      ],
    })
  })

  it('should be able to unset operator-field', () => {
    const thresholdBuilder: ThresholdEffectiveWhenBuilder = new ThresholdEffectiveWhenBuilder().use(
      {
        url: Constants.thresholdSet.effectiveWhen,
        extension: [
          {
            url: Constants.thresholdSet.operator,
            valueCode: '<=',
          },
        ],
      }
    )

    expect(thresholdBuilder.setOperator().build()).to.eql({
      url: Constants.thresholdSet.effectiveWhen,
      extension: [],
    })
  })

  it('should be able to return operator-field', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.operator,
              valueCode: '<=',
            },
          ],
        })
        .getOperator()
    ).to.eql('<=')
  })

  it('should throw error if operator-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.operator,
              valueCode: '<=',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.operator,
          valueCode: '<=',
        })
        .getOperator()
    }).to.throw(MalformedError)
  })

  it('should throw error if effectiveBehavior-extension is not one of the allowed values', () => {
    expect(() => {
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [
            {
              url: Constants.thresholdSet.operator,
              valueCode: 'something-else',
            },
          ],
        })
        .getOperator()
    }).to.throw(MalformedError)
  })

  it('should return undefined if operator-extension is not set', () => {
    expect(
      new ThresholdEffectiveWhenBuilder()
        .use({
          url: Constants.thresholdSet.effectiveWhen,
          extension: [],
        })
        .getOperator()
    ).to.be.an('undefined')
  })
})
