import { expect } from 'chai'
import { ThresholdSetBuilder, CodeableConceptBuilder } from '../src'
import sinon from 'sinon'
import { Constants } from 'global/constants'
import { MalformedError } from 'exceptions/malformed-error'
import { BackboneElement, Extension, Basic } from 'fhir'
import { ThresholdSetThresholdBuilder } from 'threshold-set-threshold-builder'

describe('ThresholdSetBuilder', () => {
  before(() => {
    // Mock make to avoid testing generated timestamp
    sinon
      .stub(ThresholdSetBuilder.prototype, '_generateIsoDate')
      .callsFake(() => '0000-00-00T00:00:00.000Z')
  })

  it('should be able to build minimal ThresholdSet', () => {
    expect(new ThresholdSetBuilder().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to accept existing ThresholdSet', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '1990-12-03T14:33:46.201Z',
          identifier: [
            {
              system: 'system',
              value: 'value',
            },
          ],
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .build()
    ).to.eql({
      resourceType: 'Basic',
      created: '1990-12-03T14:33:46.201Z',
      identifier: [
        {
          system: 'system',
          value: 'value',
        },
      ],
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to reset', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '1990-12-03T14:33:46.201Z',
        identifier: [
          {
            system: 'system',
            value: 'value',
          },
        ],
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
      }
    )
    expect(thresholdSetBuilder.use().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to set version-field', () => {
    expect(new ThresholdSetBuilder().setVersion('1.0.0').build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.version,
          valueString: '1.0.0',
        },
      ],
    })
  })

  it('should be able to unset version-field', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '0000-00-00T00:00:00.000Z',
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
        extension: [
          {
            url: Constants.thresholdSet.version,
            valueString: '1.0.0',
          },
        ],
      }
    )

    expect(thresholdSetBuilder.setVersion().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to return version-field', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.version,
              valueString: '1.0.0',
            },
          ],
        })
        .getVersion()
    ).to.eql('1.0.0')
  })

  it('should throw error if version-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.version,
              valueString: '1.0.0',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.version,
          valueString: '1.0.0',
        })
        .getVersion()
    }).to.throw(MalformedError)
  })

  it('should return undefined if version-extension is not set', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .getVersion()
    ).to.be.an('undefined')
  })

  it('should be able to set status-field', () => {
    expect(new ThresholdSetBuilder().setStatus('retired').build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.status,
          valueString: 'retired',
        },
      ],
    })
  })

  it('should be able to unset status-field', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '0000-00-00T00:00:00.000Z',
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
        extension: [
          {
            url: Constants.thresholdSet.status,
            valueString: 'retired',
          },
        ],
      }
    )

    expect(thresholdSetBuilder.setStatus().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to return status-field', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.status,
              valueString: 'retired',
            },
          ],
        })
        .getStatus()
    ).to.eql('retired')
  })

  it('should throw error if status-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.status,
              valueString: 'retired',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.status,
          valueString: 'retired',
        })
        .getStatus()
    }).to.throw(MalformedError)
  })

  it('should throw error if status-extension is not one of the allowed values', () => {
    expect(() => {
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.status,
              valueString: 'wrong-value',
            },
          ],
        })
        .getStatus()
    }).to.throw(MalformedError)
  })

  it('should return undefined if status-extension is not set', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .getStatus()
    ).to.be.an('undefined')
  })

  it('should be able to set approvalDate-field', () => {
    expect(
      new ThresholdSetBuilder()
        .setApprovalDate('2013-06-08T10:57:34+01:00')
        .build()
    ).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.approvalDate,
          valueDateTime: '2013-06-08T10:57:34+01:00',
        },
      ],
    })
  })

  it('should be able to unset approvalDate-field', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '0000-00-00T00:00:00.000Z',
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
        extension: [
          {
            url: Constants.thresholdSet.approvalDate,
            valueDateTime: '2013-06-08T10:57:34+01:00',
          },
        ],
      }
    )

    expect(thresholdSetBuilder.setApprovalDate().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to return approvalDate-field', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.approvalDate,
              valueDateTime: '2013-06-08T10:57:34+01:00',
            },
          ],
        })
        .getApprovalDate()
    ).to.eql('2013-06-08T10:57:34+01:00')
  })

  it('should throw error if approvalDate-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.approvalDate,
              valueDateTime: '2013-06-08T10:57:34+01:00',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.approvalDate,
          valueDateTime: '2013-06-08T10:57:34+01:00',
        })
        .getApprovalDate()
    }).to.throw(MalformedError)
  })

  it('should return undefined if approvalDate-extension is not set', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .getApprovalDate()
    ).to.be.an('undefined')
  })

  it('should be able to set approvingPractitioner-field', () => {
    expect(
      new ThresholdSetBuilder()
        .setApprovingPractitioner({
          reference: 'http://example.com/Practitioner/7',
          type: 'Practitioner',
        })
        .build()
    ).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.approvingPractitioner,
          valueReference: {
            reference: 'http://example.com/Practitioner/7',
            type: 'Practitioner',
          },
        },
      ],
    })
  })

  it('should be able to unset approvingPractitioner-field', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '0000-00-00T00:00:00.000Z',
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
        extension: [
          {
            url: Constants.thresholdSet.approvingPractitioner,
            valueReference: {
              reference: 'http://example.com/Practitioner/7',
              type: 'Practitioner',
            },
          },
        ],
      }
    )

    expect(thresholdSetBuilder.setApprovingPractitioner().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to return approvingPractitioner-field', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.approvingPractitioner,
              valueReference: {
                reference: 'http://example.com/Practitioner/7',
                type: 'Practitioner',
              },
            },
          ],
        })
        .getApprovingPractitioner()
    ).to.eql({
      reference: 'http://example.com/Practitioner/7',
      type: 'Practitioner',
    })
  })

  it('should throw error if approvingPractitioner-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.approvingPractitioner,
              valueReference: {
                reference: 'http://example.com/Practitioner/7',
                type: 'Practitioner',
              },
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.approvingPractitioner,
          valueReference: {
            reference: 'http://example.com/Practitioner/7',
            type: 'Practitioner',
          },
        })
        .getApprovingPractitioner()
    }).to.throw(MalformedError)
  })

  it('should return undefined if approvingPractitioner-extension is not set', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .getApprovingPractitioner()
    ).to.be.an('undefined')
  })

  it('should be able to set appliesTo-field to Reference', () => {
    expect(
      new ThresholdSetBuilder()
        .setAppliesTo({
          reference: 'http://example.com/Observation/7',
          type: 'Observation',
        })
        .build()
    ).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.appliesTo,
          valueReference: {
            reference: 'http://example.com/Observation/7',
            type: 'Observation',
          },
        },
      ],
    })
  })

  it('should be able to set appliesTo-field to canonical', () => {
    expect(
      new ThresholdSetBuilder()
        .setAppliesTo(
          'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0'
        )
        .build()
    ).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.appliesTo,
          valueCanonical:
            'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
        },
      ],
    })
  })

  it('should throw error when attempting to set malformed appliesTo', () => {
    expect(() => {
      new ThresholdSetBuilder().setAppliesTo({})
    }).to.throw(MalformedError)
  })

  it('should be able to unset appliesTo-field', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '0000-00-00T00:00:00.000Z',
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
        extension: [
          {
            url: Constants.thresholdSet.appliesTo,
            valueCanonical:
              'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
          },
        ],
      }
    )

    expect(thresholdSetBuilder.setAppliesTo().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
    })
  })

  it('should be able to return appliesTo-field', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              valueCanonical:
                'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
            },
          ],
        })
        .getAppliesTo()
    ).to.eql(
      'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0'
    )
  })

  it('should throw error if appliesTo-extension is malformed or there are more than one', () => {
    expect(() => {
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              valueCanonical:
                'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
            },
          ],
        })
        .addExtension({
          url: Constants.thresholdSet.appliesTo,
          valueCanonical:
            'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
        })
        .getAppliesTo()
    }).to.throw(MalformedError)
  })

  it('should return undefined if appliesTo-extension is not set', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '0000-00-00T00:00:00.000Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
        })
        .getAppliesTo()
    ).to.be.an('undefined')
  })

  it('should be able to set threshold-array', () => {
    const threshold: Extension = new ThresholdSetThresholdBuilder()
      .setType(new CodeableConceptBuilder().setText('text').build())
      .build()

    expect(new ThresholdSetBuilder().setThreshold([threshold]).build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.threshold,
          extension: [
            {
              url: Constants.thresholdSet.type,
              valueCodeableConcept: {
                text: 'text',
              },
            },
          ],
        },
      ],
    })
  })

  it('should be able to unset threshold-array', () => {
    const thresholdSetBuilder: ThresholdSetBuilder = new ThresholdSetBuilder().use(
      {
        resourceType: 'Basic',
        created: '0000-00-00T00:00:00.000Z',
        code: {
          coding: [
            {
              system:
                'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
              code: 'thresholdset',
              display: 'ThresholdSet',
            },
          ],
          text: 'ThresholdSet',
        },
        extension: [
          {
            url: Constants.thresholdSet.threshold,
            extension: [
              {
                url: Constants.thresholdSet.type,
                valueCodeableConcept: {
                  text: 'text',
                },
              },
            ],
          },
          {
            url: Constants.thresholdSet.version,
            valueString: '1.0.0',
          },
        ],
      }
    )

    expect(thresholdSetBuilder.setThreshold().build()).to.eql({
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.version,
          valueString: '1.0.0',
        },
      ],
    })
  })

  it('should be able to return threshold-array', () => {
    const thresholdSet: Basic = {
      resourceType: 'Basic',
      created: '0000-00-00T00:00:00.000Z',
      code: {
        coding: [
          {
            system:
              'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
            code: 'thresholdset',
            display: 'ThresholdSet',
          },
        ],
        text: 'ThresholdSet',
      },
      extension: [
        {
          url: Constants.thresholdSet.threshold,
          extension: [
            {
              url: Constants.thresholdSet.type,
              valueCodeableConcept: {
                text: 'pathological',
              },
            },
          ],
        },
        {
          url: Constants.thresholdSet.threshold,
          extension: [
            {
              url: Constants.thresholdSet.type,
              valueCodeableConcept: {
                text: 'normal',
              },
            },
          ],
        },
        {
          url: Constants.thresholdSet.version,
          valueString: '1.0.0',
        },
      ],
    }

    expect(new ThresholdSetBuilder().use(thresholdSet).getThreshold()).to.eql([
      {
        url: Constants.thresholdSet.threshold,
        extension: [
          {
            url: Constants.thresholdSet.type,
            valueCodeableConcept: {
              text: 'pathological',
            },
          },
        ],
      },
      {
        url: Constants.thresholdSet.threshold,
        extension: [
          {
            url: Constants.thresholdSet.type,
            valueCodeableConcept: {
              text: 'normal',
            },
          },
        ],
      },
    ])
  })

  it('should return empty array when threshold-array is empty', () => {
    expect(new ThresholdSetBuilder().getThreshold()).to.eql([])
  })

  it('should be able to add resource to threshold-array', () => {
    expect(
      new ThresholdSetBuilder()
        .addThreshold({
          url: Constants.thresholdSet.threshold,
          extension: [
            {
              url: Constants.thresholdSet.type,
              valueCodeableConcept: {
                text: 'normal',
              },
            },
          ],
        })
        .getThreshold()
    ).to.eql([
      {
        url: Constants.thresholdSet.threshold,
        extension: [
          {
            url: Constants.thresholdSet.type,
            valueCodeableConcept: {
              text: 'normal',
            },
          },
        ],
      },
    ])
  })

  it('should be able to tell that two thresholdSets are in conflict, if they reference the same resource', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '1990-12-03T14:33:46.201Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              valueCanonical:
                'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
            },
          ],
        })
        .conflictsWith(
          new ThresholdSetBuilder()
            .use({
              resourceType: 'Basic',
              created: '1990-12-03T14:33:46.201Z',
              code: {
                coding: [
                  {
                    system:
                      'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                    code: 'thresholdset',
                    display: 'ThresholdSet',
                  },
                ],
                text: 'ThresholdSet',
              },
              extension: [
                {
                  url: Constants.thresholdSet.appliesTo,
                  valueCanonical:
                    'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
                },
              ],
            })
            .build()
        )
    ).to.eql(true)
  })

  it('should be able to tell that two thresholdSets are in not conflict, if they reference different resources', () => {
    expect(
      new ThresholdSetBuilder()
        .use({
          resourceType: 'Basic',
          created: '1990-12-03T14:33:46.201Z',
          code: {
            coding: [
              {
                system:
                  'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                code: 'thresholdset',
                display: 'ThresholdSet',
              },
            ],
            text: 'ThresholdSet',
          },
          extension: [
            {
              url: Constants.thresholdSet.appliesTo,
              valueCanonical:
                'canonical://uuid/Questionnaire/036509c1-35e7-458e-a056-2d2111622eb0|1.0.0',
            },
          ],
        })
        .conflictsWith(
          new ThresholdSetBuilder()
            .use({
              resourceType: 'Basic',
              created: '1990-12-03T14:33:46.201Z',
              code: {
                coding: [
                  {
                    system:
                      'https://issuetracker4s.atlassian.net/wiki/spaces/FHIRReg/pages/100728935/Custom+FHIR+Resources+coding+system',
                    code: 'thresholdset',
                    display: 'ThresholdSet',
                  },
                ],
                text: 'ThresholdSet',
              },
              extension: [
                {
                  url: Constants.thresholdSet.appliesTo,
                  valueCanonical:
                    'canonical://uuid/Questionnaire/39268cb0-efb8-4e17-9b5b-9dfc7799495c|1.0.0',
                },
              ],
            })
            .build()
        )
    ).to.eql(false)
  })

  after(() => {
    sinon.reset()
  })
})
